"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
import json
import pathlib
from copy import deepcopy
from typing import TYPE_CHECKING

from hondana.scanlator_group import ScanlatorGroup
from hondana.utils import iso_to_delta, to_snake_case

if TYPE_CHECKING:
    from hondana.http import HTTPClient
    from hondana.types_.scanlator_group import GetSingleScanlationGroupResponse

PATH: pathlib.Path = pathlib.Path(__file__).parent / "payloads" / "scanlator_group.json"

PAYLOAD: GetSingleScanlationGroupResponse = json.load(PATH.open(encoding="utf-8"))
HTTP: HTTPClient = object()  # pyright: ignore[reportAssignmentType] # this is just for test purposes.


def clone_group() -> ScanlatorGroup:
    t = deepcopy(PAYLOAD)
    return ScanlatorGroup(HTTP, t["data"])


class TestScanlatorGroup:
    def test_id(self) -> None:
        group = clone_group()
        assert group.id == PAYLOAD["data"]["id"]

    def test_attributes(self) -> None:
        group = clone_group()

        for item in PAYLOAD["data"]["attributes"]:
            if item == "publishDelay":
                item = "_publish_delay"  # noqa: PLW2901  # we made this a property to allow manipulation
            assert hasattr(group, to_snake_case(item))

    def test_relationship_length(self) -> None:
        group = clone_group()

        assert group.members is not None
        assert group.leader is not None

        obj_len = len(group.members) + 1  # leader

        assert "relationships" in PAYLOAD["data"]
        assert obj_len == len(PAYLOAD["data"]["relationships"])

    def test_datetime_properties(self) -> None:
        group = clone_group()

        assert group.created_at == datetime.datetime.fromisoformat(PAYLOAD["data"]["attributes"]["createdAt"])
        assert group.updated_at == datetime.datetime.fromisoformat(PAYLOAD["data"]["attributes"]["updatedAt"])

        if group.publish_delay or PAYLOAD["data"]["attributes"]["publishDelay"]:
            assert group.publish_delay == iso_to_delta(PAYLOAD["data"]["attributes"]["publishDelay"])



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .enums import Order
from .utils import to_camel_case

if TYPE_CHECKING:
    from typing import Self


__all__ = (
    "ArtistIncludes",
    "AuthorIncludes",
    "AuthorListOrderQuery",
    "ChapterIncludes",
    "CoverArtListOrderQuery",
    "CoverIncludes",
    "CustomListIncludes",
    "FeedOrderQuery",
    "MangaDraftListOrderQuery",
    "MangaIncludes",
    "MangaListOrderQuery",
    "Order",
    "ReportListOrderQuery",
    "ScanlatorGroupIncludes",
    "ScanlatorGroupListOrderQuery",
    "SubscriptionIncludes",
    "UserListOrderQuery",
)


class _OrderQuery:
    __slots__: tuple[str, ...] = ()

    def __init__(self, **kwargs: Order) -> None:
        if not kwargs:
            msg = "You must pass valid kwargs."
            raise TypeError(msg)

        fmt: list[str] = []
        for name, value in kwargs.items():
            if name in self.__slots__:
                setattr(self, name, value.value)
            else:
                fmt.append(name)

        if fmt:
            msg = f"You have passed invalid kwargs: {', '.join(fmt)}"
            raise TypeError(msg)

    def __repr__(self) -> str:
        opt: list[str] = []
        for key in self.__slots__:
            if val := getattr(self, key, None):
                opt.append(f"{key}={val.name}")  # noqa: PERF401
        return f"<{self.__class__.__name__} {' '.join(opt)}>"

    def to_dict(self) -> dict[str, str]:
        fmt: dict[str, str] = {}
        for item in self.__slots__:
            if val := getattr(self, item, None):  # pyright: ignore[reportAssignmentType] # only Order can be assigned here
                val: str
                fmt[to_camel_case(str(item))] = val

        return fmt


class _Includes:
    __slots__: tuple[str, ...] = ()

    def to_query(self) -> list[str]:
        """Generates a list of strings based on the kwargs.

        Returns
        -------
        List[:class:`str`]
            The list of query parameters (pre-PHP formatting).
        """
        fmt: list[str] = []
        for item in dir(self):
            if item.startswith("_"):
                continue
            if getattr(self, item) and item in self.__slots__:
                fmt.append(item)

        return fmt

    def __repr__(self) -> str:
        fmt = " ".join([f"{key}={getattr(self, key, False)}" for key in self.__slots__])
        return f"<{self.__class__.__name__} {fmt}>"

    @classmethod
    def all(cls: type[Self]) -> Self:
        """A factory method that returns all possible reference expansions for this type.

        Returns
        -------
        An Includes type object with all flags set.
        """
        self = cls()

        for item in self.__slots__:
            setattr(self, item, True)

        return self

    @classmethod
    def none(cls: type[Self]) -> Self:
        """A factory method that disables all possible reference expansions for this type.

        Returns
        -------
        An Includes type object with no flags set.
        """
        self = cls()

        for item in self.__slots__:
            setattr(self, item, False)

        return self


class MangaListOrderQuery(_OrderQuery):
    """Query ordering for Manga lists.

    Parameters
    ----------
    title: :class:`~hondana.query.Order`
        Title ordering.
    year: :class:`~hondana.query.Order`
        Year ordering.
    created_at: :class:`~hondana.query.Order`
        Ordering by creation date.
    updated_at: :class:`~hondana.query.Order`
        Ordering by last updated date.
    latest_uploaded_chapter: :class:`~hondana.query.Order`
        Ordering by latest uploaded chapter.
    relevance: :class:`~hondana.query.Order`
        Ordering by relevance to search query.
    rating: :class:`~hondana.query.Order`
        Ordering by rating.
    """

    __slots__ = (
        "created_at",
        "latest_uploaded_chapter",
        "rating",
        "relevance",
        "title",
        "updated_at",
        "year",
    )

    title: Order | None
    year: Order | None
    created_at: Order | None
    updated_at: Order | None
    latest_uploaded_chapter: Order | None
    relevance: Order | None
    rating: Order | None

    def __init__(self, **kwargs: Order) -> None:
        super().__init__(**kwargs)


class FeedOrderQuery(_OrderQuery):
    """Query ordering for feeds.

    Parameters
    ----------
    created_at: :class:`~hondana.query.Order`
        Ordering by creation date.
    updated_at: :class:`~hondana.query.Order`
        Ordering by last updated date.
    publish_at: :class:`~hondana.query.Order`
        Ordering by published at date.
    readable_at: :class:`~hondana.query.Order`
        Ordering by when the chapters are readable.
    volume: :class:`~hondana.query.Order`
        Ordering by volume number.
    chapter: :class:`~hondana.query.Order`
        Ordering by chapter number.
    """

    __slots__ = (
        "chapter",
        "created_at",
        "publish_at",
        "readable_at",
        "updated_at",
        "volume",
    )

    created_at: Order | None
    updated_at: Order | None
    publish_at: Order | None
    readable_at: Order | None
    volume: Order | None
    chapter: Order | None

    def __init__(self, **kwargs: Order) -> None:
        super().__init__(**kwargs)


class MangaDraftListOrderQuery(_OrderQuery):
    """Query ordering for Manga draft lists.

    Parameters
    ----------
    title: :class:`~hondana.query.Order`
        Title ordering.
    year: :class:`~hondana.query.Order`
        Year ordering.
    created_at: :class:`~hondana.query.Order`
        Ordering by creation date.
    updated_at: :class:`~hondana.query.Order`
        Ordering by last updated date.
    """

    __slots__ = (
        "created_at",
        "title",
        "updated_at",
        "year",
    )

    title: Order | None
    year: Order | None
    created_at: Order | None
    updated_at: Order | None

    def __init__(self, **kwargs: Order) -> None:
        super().__init__(**kwargs)


class CoverArtListOrderQuery(_OrderQuery):
    """Query ordering for Cover art lists.

    Parameters
    ----------
    created_at: :class:`~hondana.query.Order`
        Ordering by creation date.
    updated_at: :class:`~hondana.query.Order`
        Ordering by last updated date.
    volume: :class:`~hondana.query.Order`
        Ordering by volume number.
    """

    __slots__ = (
        "created_at",
        "updated_at",
        "volume",
    )

    created_at: Order | None
    updated_at: Order | None
    volume: Order | None

    def __init__(self, **kwargs: Order) -> None:
        super().__init__(**kwargs)


class ScanlatorGroupListOrderQuery(_OrderQuery):
    """Query ordering for Scanlator group lists.

    Parameters
    ----------
    name: :class:`~hondana.query.Order`
        Name ordering.
    created_at: :class:`~hondana.query.Order`
        Ordering by creation date.
    updated_at: :class:`~hondana.query.Order`
        Ordering by last updated date.
    followed_count: :class:`~hondana.query.Order`
        Ordering by followed count.
    relevance: :class:`~hondana.query.Order`
        Ordering by relevance to search query.
    """

    __slots__ = (
        "created_at",
        "followed_count",
        "name",
        "relevance",
        "updated_at",
    )

    name: Order | None
    created_at: Order | None
    updated_at: Order | None
    followed_count: Order | None
    relevance: Order | None

    def __init__(self, **kwargs: Order) -> None:
        super().__init__(**kwargs)


class AuthorListOrderQuery(_OrderQuery):
    """Query ordering for Author lists.

    Parameters
    ----------
    name: :class:`~hondana.query.Order`
        Name ordering.
    """

    __slots__ = ("name",)

    name: Order | None

    def __init__(self, **kwargs: Order) -> None:
        super().__init__(**kwargs)


class UserListOrderQuery(_OrderQuery):
    """Query ordering for User lists.

    Parameters
    ----------
    username: :class:`~hondana.query.Order`
        Username ordering.
    """

    __slots__ = ("username",)

    username: Order | None

    def __init__(self, **kwargs: Order) -> None:
        super().__init__(**kwargs)


class ReportListOrderQuery(_OrderQuery):
    """Query ordering for Report lists.

    Parameters
    ----------
    created_at: :class:`~hondana.query.Order`
        Ordering by creation date.
    """

    __slots__ = ("created_at",)

    created_at: Order | None

    def __init__(self, **kwargs: Order) -> None:
        super().__init__(**kwargs)


class ArtistIncludes(_Includes):
    """
    A helper for generating the ``includes[]`` parameter for queries.

    Parameters
    ----------
    manga: :class:`bool`
        Defaults to ``True``. Whether to include manga in the relationships.
    """

    __slots__ = ("manga",)

    def __init__(self, *, manga: bool = True) -> None:
        self.manga: bool = manga

    def to_query(self) -> list[str]:
        """Generates a list of strings based on the kwargs.

        Returns
        -------
        List[:class:`str`]
            The list of query parameters (pre-PHP formatting).
        """
        return super().to_query()


class AuthorIncludes(_Includes):
    """
    A helper for generating the ``includes[]`` parameter for queries.

    Parameters
    ----------
    manga: :class:`bool`
        Defaults to ``True``. Whether to include manga in the relationships.
    """

    __slots__ = ("manga",)

    def __init__(self, *, manga: bool = True) -> None:
        self.manga: bool = manga

    def to_query(self) -> list[str]:
        """Generates a list of strings based on the kwargs.

        Returns
        -------
        List[:class:`str`]
            The list of query parameters (pre-PHP formatting).
        """
        return super().to_query()


class ChapterIncludes(_Includes):
    """
    A helper for generating the ``includes[]`` parameter for queries.

    Parameters
    ----------
    manga: :class:`bool`
        Defaults to ``True``. Whether to include manga in the relationships.
    user: :class:`bool`
        Defaults to ``True``. Whether to include user in the relationships.
    scanlation_group: :class:`bool`
        Defaults to ``True``. Whether to include scanlator group in the relationships.
    """

    __slots__ = (
        "manga",
        "scanlation_group",
        "user",
    )

    def __init__(self, *, manga: bool = True, user: bool = True, scanlation_group: bool = True) -> None:
        self.manga: bool = manga
        self.user: bool = user
        self.scanlation_group: bool = scanlation_group

    def to_query(self) -> list[str]:
        """Returns a list of valid query strings.

        Returns
        -------
        List[:class:`str`]
            The list of query parameters (pre-PHP formatting).
        """
        return super().to_query()


class CoverIncludes(_Includes):
    """
    A helper for generating the ``includes[]`` parameter for queries.

    Parameters
    ----------
    manga: :class:`bool`
        Defaults to ``True``. Whether to include manga in the relationships.
    user: :class:`bool`
        Defaults to ``True``. Whether to include user in the relationships.
    """

    __slots__ = (
        "manga",
        "user",
    )

    def __init__(self, *, manga: bool = True, user: bool = True) -> None:
        self.manga: bool = manga
        self.user: bool = user

    def to_query(self) -> list[str]:
        """Returns a list of valid query strings.

        Returns
        -------
        List[:class:`str`]
            The list of query parameters (pre-PHP formatting).
        """
        return super().to_query()


class CustomListIncludes(_Includes):
    """
    A helper for generating the ``includes[]`` parameter for queries.

    Parameters
    ----------
    manga: :class:`bool`
        Defaults to ``True``. Whether to include manga in the relationships.
    user: :class:`bool`
        Defaults to ``True``. Whether to include user in the relationships.
    owner: :class:`bool`
        Defaults to ``True``. Whether to include owner in the relationships.
    """

    __slots__ = (
        "manga",
        "owner",
        "user",
    )

    def __init__(self, *, manga: bool = True, user: bool = True, owner: bool = True) -> None:
        self.manga: bool = manga
        self.user: bool = user
        self.owner: bool = owner

    def to_query(self) -> list[str]:
        """Returns a list of valid query strings.

        Returns
        -------
        List[:class:`str`]
            The formatted query string.
        """
        return super().to_query()


class MangaIncludes(_Includes):
    """
    A helper for generating the ``includes[]`` parameter for queries.

    Parameters
    ----------
    author: :class:`bool`
        Defaults to ``True``. Whether to include author in the relationships.
    artist: :class:`bool`
        Defaults to ``True``. Whether to include artist in the relationships.
    cover_art: :class:`bool`
        Defaults to ``True``. Whether to include cover in the relationships.
    manga: :class:`bool`
        Defaults to ``True``. Whether to include manga in the relationships.
    """

    __slots__ = (
        "artist",
        "author",
        "cover_art",
        "manga",
    )

    def __init__(self, *, author: bool = True, artist: bool = True, cover_art: bool = True, manga: bool = True) -> None:
        self.author: bool = author
        self.artist: bool = artist
        self.cover_art: bool = cover_art
        self.manga: bool = manga


class ScanlatorGroupIncludes(_Includes):
    """
    A helper for generating the ``includes[]`` parameter for queries.

    Parameters
    ----------
    leader: :class:`bool`
        Defaults to ``True``. Whether to include leader in the relationships.
    member: :class:`bool`
        Defaults to ``True``. Whether to include members in the relationships.
    """

    __slots__ = (
        "leader",
        "member",
    )

    def __init__(self, *, leader: bool = True, member: bool = True) -> None:
        self.leader: bool = leader
        self.member: bool = member

    def to_query(self) -> list[str]:
        """Returns a list of valid query strings.

        Returns
        -------
        List[:class:`str`]
            The list of query parameters (pre-PHP formatting).
        """
        return super().to_query()


class UserReportIncludes(_Includes):
    """
    A helper for generating the ``includes[]`` parameter for queries.

    Parameters
    ----------
    user: :class:`bool`
        Defaults to ``True``. Whether to include user in the relationships.
    reason: :class:`bool`
        Defaults to ``True``. Whether to include report in the relationships.
    """

    __slots__ = (
        "reason",
        "user",
    )

    def __init__(self, *, user: bool = True, reason: bool = True) -> None:
        self.user: bool = user
        self.reason: bool = reason

    def to_query(self) -> list[str]:
        """Returns a list of valid query strings.

        Returns
        -------
        List[:class:`str`]
            The list of query parameters (pre-PHP formatting).
        """
        return super().to_query()


class SubscriptionIncludes(_Includes):
    """
    A helper for generating the ``includes[]`` parameter for queries.

    Parameters
    ----------
    user: :class:`bool`
        Defaults to ``True``. Whether to include user in the relationships.
    """

    __slots__ = ("user",)

    def __init__(self, *, user: bool = True) -> None:
        self.user: bool = user

    def to_query(self) -> list[str]:
        """Returns a list of valid query strings.

        Returns
        -------
        List[:class:`str`]
            The list of query parameters (pre-PHP formatting).
        """
        return super().to_query()



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import aiohttp

    from .types_.errors import ErrorType

__all__ = (
    "APIException",
    "AuthenticationRequired",
    "BadRequest",
    "Forbidden",
    "MangaDexServerError",
    "NotFound",
    "PreviousAPIVersionRequest",
    "RefreshTokenFailure",
    "TermsOfServiceNotAccepted",
    "Unauthorized",
    "UploadInProgress",
)


class Error:
    """
    Common Error type class representing an error returned from MangaDex.

    Attributes
    ----------
    error_id: :class:`str`
        The UUID representing this error.
    error_status: :class:`int`
        The HTTP status code of this error.
    error_title: :class:`str`
        The title summary of this error.
    error_detail: :class:`str`
        The detailed report of this error.
    """

    __slots__ = (
        "error_detail",
        "error_id",
        "error_status",
        "error_title",
    )

    def __init__(self, error: ErrorType) -> None:
        self.error_id: str = error["id"]
        self.error_status: int = error["status"]
        self.error_title: str = error["title"]
        self.error_detail: str = error["detail"]

    def __repr__(self) -> str:
        return f"<Error status={self.error_status} title={self.error_title!r} detail={self.error_detail!r}>"

    def __str__(self) -> str:
        return f"Error: {self.error_title}: {self.error_detail}"


class AuthenticationRequired(Exception):
    """An exception to be raised when authentication is required to use this endpoint."""


class RefreshTokenFailure(Exception):
    """An exception to be raised when trying to use or access the refresh token fails.

    Attriutes
    ----------
    message: :class:`str`
        The error message.
    response: :class:`aiohttp.ClientResponse`
        The client response the error originated from.
    data: Dict[:class:`str`, :class:`~typing.Any`]
        The data from the error.
    """

    def __init__(self, message: str, /, response: aiohttp.ClientResponse, response_data: dict[str, Any]) -> None:
        self.message = message
        self.response = response
        self.data = response_data

    @property
    def status(self) -> int:
        """The HTTP response code from the error we received.

        Returns
        -------
        :class:`int`
        """
        return self.response.status


class UploadInProgress(Exception):
    """An exception to be raised when an upload in progress is already found for the logged-in user.

    Attributes
    ----------
    message: class:`str`
        The message returned with this error.
    session_id: :class:`str`
        The already existing session ID.
    """

    __slots__ = (
        "message",
        "session_id",
    )

    def __init__(self, message: str, /, *, session_id: str) -> None:
        self.message: str = message
        self.session_id: str = session_id
        super().__init__(self.message, self.session_id)

    def __str__(self) -> str:
        return f"An upload session was already found, it's ID is: {self.session_id}"


class MangaDexServerError(Exception):
    """Generic exception type for when MangaDex is down.

    Attributes
    ----------
    response: :class:`aiohttp.ClientResponse`
        The response object pertaining to this request.
    status_code: :class:`int`
        The HTTP status code for this request.
    """

    __slots__ = (
        "response",
        "status_code",
    )

    def __init__(self, response: aiohttp.ClientResponse, /, *, status_code: int) -> None:
        self.response: aiohttp.ClientResponse = response
        self.status_code: int = status_code
        super().__init__(self.response, self.status_code)


class APIException(Exception):
    """Generic API error when the response code is a non 2xx error.

    Attributes
    ----------
    response: :class:`aiohttp.ClientResponse`
        The response object pertaining to this request.
    status_code: :class:`int`
        The HTTP status code for this request.
    response_id: :class:`str`
        The UUID relating to this failed HTTP request.
        This is to be used when contacting MangaDex about an error.
    errors: List[:class:`~hondana.errors.Error`]
        The list of errors returned from MangaDex.
    """

    __slots__ = (
        "_errors",
        "errors",
        "response",
        "response_id",
        "status_code",
    )

    def __init__(self, response: aiohttp.ClientResponse, /, *, status_code: int, errors: list[ErrorType]) -> None:
        self.response: aiohttp.ClientResponse = response
        self.status_code: int = status_code
        self.response_id: str = response.headers["x-request-id"]
        self._errors: list[ErrorType] = errors
        self.errors: list[Error] = []
        self.errors.extend(Error(item) for item in self._errors)
        super().__init__(self.status_code, self.errors)

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"response_id={self.response_id} "
            f"status_code={self.status_code} "
            f"error_amount: {len(self.errors)}>"
        )

    def __str__(self) -> str:
        return (
            f"HTTP Status: {self.status_code} and response id: {self.response_id} :: "
            f"{', '.join([error.error_detail for error in self.errors])}"
        )


class BadRequest(APIException):
    """An error for when the API query was malformed.

    Attributes
    ----------
    response: :class:`aiohttp.ClientResponse`
        The response object pertaining to this request.
    status_code: Literal[``400``]
        The HTTP status code for this request.
    response_id: :class:`str`
        The UUID relating to this failed HTTP request.
        This is to be used when contacting MangaDex about an error.
    errors: List[:class:`~hondana.errors.Error`]
        The list of errors returned from MangaDex.
    """

    def __init__(self, response: aiohttp.ClientResponse, /, *, errors: list[ErrorType]) -> None:
        super().__init__(response, status_code=400, errors=errors)


class PreviousAPIVersionRequest(BadRequest):
    """
    An error for when the API query matches the criteria of an older API version.

    Attributes
    ----------
    response: :class:`aiohttp.ClientResponse`
        The response object pertaining to this request.
    status_code: Literal[``400``]
        The HTTP status code for this request.
    response_id: :class:`str`
        The UUID relating to this failed HTTP request.
        This is to be used when contacting MangaDex about an error.
    errors: List[:class:`~hondana.errors.Error`]
        The list of errors returned from MangaDex.
    """

    def __init__(self, response: aiohttp.ClientResponse, /) -> None:
        super().__init__(response, errors=[])

    def __str__(self) -> str:
        return (
            f"HTTP Status: {self.status_code} and response id: {self.response_id} :: "
            "Your API request matches the critera for MangaDex API version <5."
        )


class Unauthorized(APIException):
    """An error for when you are unauthorized on this API endpoint.

    Attributes
    ----------
    response: :class:`aiohttp.ClientResponse`
        The response object pertaining to this request.
    status_code: Literal[``401``]
        The HTTP status code for this request.
    response_id: :class:`str`
        The UUID relating to this failed HTTP request.
        This is to be used when contacting MangaDex about an error.
    errors: List[:class:`~hondana.errors.Error`]
        The list of errors returned from MangaDex.
    """

    def __init__(self, response: aiohttp.ClientResponse, /, *, errors: list[ErrorType]) -> None:
        super().__init__(response, status_code=401, errors=errors)


class Forbidden(APIException):
    """An error for when your authorization was rejected.

    Attributes
    ----------
    response: :class:`aiohttp.ClientResponse`
        The response object pertaining to this request.
    status_code: Literal[``403``]
        The HTTP status code for this request.
    response_id: :class:`str`
        The UUID relating to this failed HTTP request.
        This is to be used when contacting MangaDex about an error.
    errors: List[:class:`~hondana.errors.Error`]
        The list of errors returned from MangaDex.
    """

    def __init__(self, response: aiohttp.ClientResponse, /, *, errors: list[ErrorType]) -> None:
        super().__init__(response, status_code=403, errors=errors)


class NotFound(APIException):
    """An error for when the requested API item was not found."

    Attributes
    ----------
    response: :class:`aiohttp.ClientResponse`
        The response object pertaining to this request.
    status_code: Literal[``404``]
        The HTTP status code for this request.
    response_id: :class:`str`
        The UUID relating to this failed HTTP request.
        This is to be used when contacting MangaDex about an error.
    errors: List[:class:`~hondana.errors.Error`]
        The list of errors returned from MangaDex.
    """

    def __init__(self, response: aiohttp.ClientResponse, /, *, errors: list[ErrorType]) -> None:
        super().__init__(response, status_code=404, errors=errors)


class TermsOfServiceNotAccepted(Exception):
    """An error for when the ToS is not accepted prior to a chapter upload."""



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
import json
import pathlib
from copy import deepcopy
from typing import TYPE_CHECKING

from hondana.artist import Artist
from hondana.utils import RelationshipResolver, to_snake_case

if TYPE_CHECKING:
    from hondana.http import HTTPClient
    from hondana.types_.artist import GetSingleArtistResponse
    from hondana.types_.manga import MangaResponse

PATH: pathlib.Path = pathlib.Path(__file__).parent / "payloads" / "artist.json"


PAYLOAD: GetSingleArtistResponse = json.load(PATH.open(encoding="utf-8"))
HTTP: HTTPClient = object()  # pyright: ignore[reportAssignmentType] # this is just for test purposes.


def clone_artist() -> Artist:
    t = deepcopy(PAYLOAD)
    return Artist(HTTP, t["data"])


class TestArtist:
    def test_id(self) -> None:
        artist = clone_artist()
        assert artist.id == PAYLOAD["data"]["id"]

    def test_attributes(self) -> None:
        artist = clone_artist()
        for item in PAYLOAD["data"]["attributes"]:
            getattr(artist, to_snake_case(item))

    def test_relationship_length(self) -> None:
        artist = clone_artist()
        obj_len = len(artist._manga_relationships)  # pyright: ignore[reportPrivateUsage] # sorry, need this for test purposes
        assert "relationships" in PAYLOAD["data"]

        # remove empty relationships
        resolved_relationships = [r for r in PAYLOAD["data"]["relationships"] if "attributes" in r]

        assert obj_len == len(resolved_relationships)

    def test_manga_relationships(self) -> None:
        artist = clone_artist()
        assert "relationships" in PAYLOAD["data"]
        rels = RelationshipResolver["MangaResponse"](PAYLOAD["data"]["relationships"], "manga").resolve(remove_empty=True)

        assert artist.manga is not None
        assert len(rels) == len(artist.manga)

    def test_datetime_props(self) -> None:
        artist = clone_artist()

        assert artist.created_at == datetime.datetime.fromisoformat(PAYLOAD["data"]["attributes"]["createdAt"])
        assert artist.updated_at == datetime.datetime.fromisoformat(PAYLOAD["data"]["attributes"]["updatedAt"])



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
import pathlib
import random
import zoneinfo
from typing import TYPE_CHECKING, TypeVar

import pytest
from multidict import MultiDict

from hondana.utils import (
    MISSING,
    RelationshipResolver,
    Route,
    as_chunks,
    calculate_limits,
    clean_isoformat,
    delta_to_iso,
    iso_to_delta,
    php_query_builder,
    to_camel_case,
    to_snake_case,
    upload_file_sort,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from hondana.utils import MANGADEX_QUERY_PARAM_TYPE


T = TypeVar("T")


multidict_one = MultiDict[str]()
multidict_one.add("order[publishAt]", "desc")
multidict_one.add("translatedLanguages[]", "en")
multidict_one.add("translatedLanguages[]", "jp")

multidict_two = MultiDict[str]()
multidict_two.add("param[]", "a")
multidict_two.add("param[]", "b")
multidict_two.add("param[]", "c")
multidict_two.add("includes[]", "test")
multidict_two.add("includes[]", "mark")

multidict_three = MultiDict[str | int]()
multidict_three.add("sorted", "true")
multidict_three.add("value", 1)

multidict_four = MultiDict[str | int]()
multidict_four.add("sorted[]", 1)
multidict_four.add("sorted[]", 2)
multidict_four.add("includes[]", 1)
multidict_four.add("includes[]", 2)
multidict_four.add("includes[]", 3)
multidict_four.add("createdAt", "null")
multidict_four.add("value[query]", 7)


class TestUtils:
    def test_missing(self) -> None:
        assert MISSING is not None
        assert not bool(MISSING)

    @pytest.mark.parametrize(
        "route, match",
        [
            (
                Route("GET", "/chapter/{chapter_id}", chapter_id="abcd", authenticate=False),
                ("GET", "/chapter/abcd", False),
            ),
            (
                Route("POST", "/manga/{manga_id}", manga_id="some_manga", authenticate=True),
                ("POST", "/manga/some_manga", True),
            ),
        ],
    )
    def test_route(self, route: Route, match: tuple[str, str, bool]) -> None:
        verb, path, auth = match

        assert route.verb == verb

        assert route.url.scheme == "https"
        assert route.url.host == "api.mangadex.org"
        assert route.url.path == path

        assert route.auth is auth

    @pytest.mark.parametrize(
        "source, chunk_size, chunked",
        [
            ([1, 2, 3, 4, 5, 6], 2, [[1, 2], [3, 4], [5, 6]]),
            ([1, 2, 3, 4, 5, 6], 3, [[1, 2, 3], [4, 5, 6]]),
            ([1, 2, 3, 4, 5, 6], 4, [[1, 2, 3, 4], [5, 6]]),
            ([1, 2, 3, 4, 5, 6], 5, [[1, 2, 3, 4, 5], [6]]),
        ],
    )
    def test_as_chunks(self, source: Iterable[T], chunk_size: int, chunked: Iterable[Iterable[T]]) -> None:
        assert list(as_chunks(source, chunk_size)) == chunked

    @pytest.mark.parametrize(
        "limit, offset, max_limit, output",
        [
            (100, 9980, 100, (20, 9980)),
            (10, 9999, 100, (1, 9999)),
            (100, 5000, 100, (100, 5000)),
            (100, 5000, 10, (10, 5000)),
        ],
    )
    def test_calculate_limit(self, limit: int, offset: int, max_limit: int, output: tuple[int, int]) -> None:
        assert calculate_limits(limit, offset, max_limit=max_limit) == output

    @pytest.mark.parametrize(
        "input_, output",
        [
            (datetime.timedelta(hours=3, minutes=12), "T3H12M"),
            (datetime.timedelta(days=12, hours=13, minutes=27, seconds=12), "P5D1WT13H27M12S"),
            (datetime.timedelta(days=7, hours=7, minutes=13, seconds=58), "P1WT7H13M58S"),
        ],
    )
    def test_delta_to_iso(self, input_: datetime.timedelta, output: str) -> None:
        assert delta_to_iso(input_) == output

    @pytest.mark.parametrize(
        "input_, output",
        [
            ("T3H12M", datetime.timedelta(seconds=11520)),
            ("P5D1WT13H27M12S", datetime.timedelta(days=12, seconds=48432)),
            ("P1WT7H13M58S", datetime.timedelta(days=7, seconds=26038)),
        ],
    )
    def test_iso_to_delta(self, input_: str, output: datetime.timedelta) -> None:
        assert iso_to_delta(input_) == output

    @pytest.mark.parametrize(
        "input_, output",
        [
            (
                {"order": {"publishAt": "desc"}, "translatedLanguages": ["en", "jp"]},
                multidict_one,
            ),
            (
                {"param": ["a", "b", "c"], "includes": ["test", "mark"]},
                multidict_two,
            ),
            ({"sorted": True, "value": 1}, multidict_three),
            (
                {"sorted": [1, 2], "includes": [1, 2, 3], "createdAt": None, "value": {"query": 7}},
                multidict_four,
            ),
        ],
    )
    def test_query_builder(
        self,
        input_: MANGADEX_QUERY_PARAM_TYPE,
        output: MultiDict[str | int],
    ) -> None:
        assert php_query_builder(input_) == output

    @pytest.mark.parametrize(
        "input_, output",
        [("some_value", "someValue"), ("some_other_value", "someOtherValue"), ("manga_or_chapter", "mangaOrChapter")],
    )
    def test_to_camel_case(self, input_: str, output: str) -> None:
        assert to_camel_case(input_) == output

    @pytest.mark.parametrize(
        "input_, output",
        [("someValue", "some_value"), ("someOtherValue", "some_other_value"), ("mangaOrChapter", "manga_or_chapter")],
    )
    def test_to_snake_case(self, input_: str, output: str) -> None:
        assert to_snake_case(input_) == output

    @pytest.mark.parametrize(
        "input_, output",
        [
            (
                [{"type": "test", "hello": "goodbye"}],
                [{"type": "test", "hello": "goodbye"}],
            ),
            (
                [{"type": "bleh", "hello": "goodbye"}, {"type": "test", "hello": "hello"}],
                [{"type": "test", "hello": "hello"}],
            ),
            (
                [{"type": "test", "pass": "mark", "something": "anything"}, {"type": "bad"}],
                [{"type": "test", "pass": "mark", "something": "anything"}],
            ),
        ],
    )
    def test_relationship_finder(self, input_: list[dict[str, str]], output: list[dict[str, str]] | dict[str, str]) -> None:
        ret = RelationshipResolver[dict[str, str]](input_, "test").resolve()  # type: ignore[reportArgumentType] # we lie here for the test case

        assert ret is not None

        assert ret == output

    @pytest.mark.parametrize(
        "input_, output",
        [
            (
                datetime.datetime(
                    year=2022,
                    month=3,
                    day=19,
                    hour=12,
                    minute=0,
                    second=0,
                    microsecond=0,
                    tzinfo=datetime.UTC,
                ),
                "2022-03-19T12:00:00",
            ),
            (
                datetime.datetime(
                    year=2022,
                    month=3,
                    day=19,
                    hour=12,
                    minute=0,
                    second=0,
                    microsecond=0,
                    tzinfo=zoneinfo.ZoneInfo("Asia/Tokyo"),
                ),
                "2022-03-19T03:00:00",
            ),
        ],
    )
    def test_isoformatter(self, input_: datetime.datetime, output: str) -> None:
        assert clean_isoformat(input_) == output

    @pytest.mark.parametrize(
        "input_, output",
        [
            (
                random.sample([pathlib.Path(f"{x}.png") for x in range(1, 15)], 14),
                [
                    pathlib.Path("1.png"),
                    pathlib.Path("2.png"),
                    pathlib.Path("3.png"),
                    pathlib.Path("4.png"),
                    pathlib.Path("5.png"),
                    pathlib.Path("6.png"),
                    pathlib.Path("7.png"),
                    pathlib.Path("8.png"),
                    pathlib.Path("9.png"),
                    pathlib.Path("10.png"),
                    pathlib.Path("11.png"),
                    pathlib.Path("12.png"),
                    pathlib.Path("13.png"),
                    pathlib.Path("14.png"),
                ],
            ),
            (
                [pathlib.Path("1.png"), pathlib.Path("11.png"), pathlib.Path("2.png"), pathlib.Path("22.png")],
                [pathlib.Path("1.png"), pathlib.Path("2.png"), pathlib.Path("11.png"), pathlib.Path("22.png")],
            ),
        ],
    )
    def test_path_sorter(self, input_: list[pathlib.Path], output: list[pathlib.Path]) -> None:
        assert sorted(input_, key=upload_file_sort) == output



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from hondana.query import (
    ArtistIncludes,
    AuthorIncludes,
    AuthorListOrderQuery,
    ChapterIncludes,
    CoverArtListOrderQuery,
    CoverIncludes,
    CustomListIncludes,
    FeedOrderQuery,
    MangaDraftListOrderQuery,
    MangaIncludes,
    MangaListOrderQuery,
    Order,
    ReportListOrderQuery,
    ScanlatorGroupIncludes,
    ScanlatorGroupListOrderQuery,
    UserListOrderQuery,
)


class TestOrderQuery:
    def test_author_list_order_query(self) -> None:
        query = AuthorListOrderQuery(name=Order.descending)
        assert query.to_dict() == {"name": "desc"}

    def test_cover_art_order_query(self) -> None:
        query = CoverArtListOrderQuery(created_at=Order.ascending)
        assert query.to_dict() == {"createdAt": "asc"}

        query = CoverArtListOrderQuery(updated_at=Order.descending)
        assert query.to_dict() == {"updatedAt": "desc"}

        query = CoverArtListOrderQuery(volume=Order.ascending)
        assert query.to_dict() == {"volume": "asc"}

    def test_feed_order_query(self) -> None:
        query = FeedOrderQuery(created_at=Order.ascending)
        assert query.to_dict() == {"createdAt": "asc"}

        query = FeedOrderQuery(updated_at=Order.ascending)
        assert query.to_dict() == {"updatedAt": "asc"}

        query = FeedOrderQuery(publish_at=Order.ascending)
        assert query.to_dict() == {"publishAt": "asc"}

        query = FeedOrderQuery(readable_at=Order.ascending)
        assert query.to_dict() == {"readableAt": "asc"}

        query = FeedOrderQuery(volume=Order.ascending)
        assert query.to_dict() == {"volume": "asc"}

        query = FeedOrderQuery(chapter=Order.ascending)
        assert query.to_dict() == {"chapter": "asc"}

    def test_manga_draft_list_order_query(self) -> None:
        query = MangaDraftListOrderQuery(title=Order.descending)
        assert query.to_dict() == {"title": "desc"}

        query = MangaDraftListOrderQuery(year=Order.descending)
        assert query.to_dict() == {"year": "desc"}

        query = MangaDraftListOrderQuery(created_at=Order.descending)
        assert query.to_dict() == {"createdAt": "desc"}

        query = MangaDraftListOrderQuery(updated_at=Order.descending)
        assert query.to_dict() == {"updatedAt": "desc"}

    def test_manga_list_order_query(self) -> None:
        query = MangaListOrderQuery(created_at=Order.descending)
        assert query.to_dict() == {"createdAt": "desc"}

        query = MangaListOrderQuery(title=Order.ascending)
        assert query.to_dict() == {"title": "asc"}

    def test_report_list_order_query(self) -> None:
        query = ReportListOrderQuery(created_at=Order.descending)
        assert query.to_dict() == {"createdAt": "desc"}

    def test_scanlator_group_list_order_query(self) -> None:
        query = ScanlatorGroupListOrderQuery(name=Order.descending)
        assert query.to_dict() == {"name": "desc"}

        query = ScanlatorGroupListOrderQuery(created_at=Order.descending)
        assert query.to_dict() == {"createdAt": "desc"}

        query = ScanlatorGroupListOrderQuery(updated_at=Order.descending)
        assert query.to_dict() == {"updatedAt": "desc"}

        query = ScanlatorGroupListOrderQuery(followed_count=Order.descending)
        assert query.to_dict() == {"followedCount": "desc"}

        query = ScanlatorGroupListOrderQuery(relevance=Order.descending)
        assert query.to_dict() == {"relevance": "desc"}

    def test_user_list_order_query(self) -> None:
        query = UserListOrderQuery(username=Order.descending)
        assert query.to_dict() == {"username": "desc"}


class TestIncludes:
    def test_artist_includes(self) -> None:
        includes = ArtistIncludes(manga=False)
        assert includes.to_query() == []

        includes = ArtistIncludes(manga=True)
        assert includes.to_query() == ["manga"]

        includes = ArtistIncludes()
        assert includes.to_query() == ArtistIncludes.all().to_query()

    def test_author_includes(self) -> None:
        includes = AuthorIncludes(manga=False)
        assert includes.to_query() == []

        includes = AuthorIncludes(manga=True)
        assert includes.to_query() == ["manga"]

        includes = AuthorIncludes()
        assert includes.to_query() == AuthorIncludes.all().to_query()

    def test_chapter_includes(self) -> None:
        includes = ChapterIncludes(manga=False, user=False, scanlation_group=False)
        assert includes.to_query() == []

        includes = ChapterIncludes(manga=True, user=False, scanlation_group=False)
        assert includes.to_query() == ["manga"]

        includes = ChapterIncludes(manga=True, user=False, scanlation_group=True)
        assert includes.to_query() == ["manga", "scanlation_group"]

        includes = ChapterIncludes()
        assert includes.to_query() == ChapterIncludes.all().to_query()

    def test_cover_includes(self) -> None:
        includes = CoverIncludes.none()
        assert includes.to_query() == []

        includes = CoverIncludes(manga=False, user=False)
        assert includes.to_query() == []

        includes = CoverIncludes(manga=True, user=False)
        assert includes.to_query() == ["manga"]

        includes = CoverIncludes()
        assert includes.to_query() == CoverIncludes.all().to_query()

    def test_custom_list_includes(self) -> None:
        includes = CustomListIncludes.none()
        assert includes.to_query() == []

        includes = CustomListIncludes()
        assert includes.to_query() == ["manga", "owner", "user"]

        includes = CustomListIncludes(manga=False)
        assert includes.to_query() == ["owner", "user"]

    def test_manga_includes(self) -> None:
        includes = MangaIncludes()
        assert includes.to_query() == ["artist", "author", "cover_art", "manga"]

        includes = MangaIncludes.none()
        assert includes.to_query() == []

        includes = MangaIncludes(artist=False)
        assert includes.to_query() == ["author", "cover_art", "manga"]

    def test_scanlator_group_includes(self) -> None:
        includes = ScanlatorGroupIncludes.none()
        assert includes.to_query() == []

        includes = ScanlatorGroupIncludes()
        assert includes.to_query() == ["leader", "member"]

        includes = ScanlatorGroupIncludes(leader=False)
        assert includes.to_query() == ["member"]



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import json
import pathlib
from typing import TYPE_CHECKING, Any, Literal, overload

from hondana.author import Author
from hondana.chapter import Chapter
from hondana.collections import (
    AuthorCollection,
    BaseCollection,
    ChapterFeed,
    CoverCollection,
    CustomListCollection,
    LegacyMappingCollection,
    MangaCollection,
    MangaRelationCollection,
    ReportCollection,
    ScanlatorGroupCollection,
    UserCollection,
    UserReportCollection,
)
from hondana.cover import Cover
from hondana.custom_list import CustomList
from hondana.legacy import LegacyItem
from hondana.manga import Manga, MangaRelation
from hondana.scanlator_group import ScanlatorGroup
from hondana.user import User

if TYPE_CHECKING:
    from hondana.http import HTTPClient
    from hondana.types_.author import GetMultiAuthorResponse
    from hondana.types_.chapter import GetMultiChapterResponse
    from hondana.types_.cover import GetMultiCoverResponse
    from hondana.types_.custom_list import GetMultiCustomListResponse
    from hondana.types_.legacy import GetLegacyMappingResponse
    from hondana.types_.manga import MangaRelationResponse, MangaSearchResponse
    from hondana.types_.scanlator_group import GetMultiScanlationGroupResponse
    from hondana.types_.user import GetMultiUserResponse

PATH: pathlib.Path = pathlib.Path(__file__).parent / "payloads" / "collections"
HTTP: HTTPClient = object()  # pyright: ignore[reportAssignmentType] # this is just for test purposes.

CollectionType = Literal[
    "authors",
    "chapter_feed",
    "covers",
    "custom_lists",
    "legacy_mapping",
    "manga",
    "manga_relations",
    "reports",  # not supported, needs moderator access
    "scanlator_groups",
    "users",  # needs authentication access
    "user_reports",  # not supported, needs moderator access
]


@overload
def clone_collection(type_: Literal["authors"], /) -> AuthorCollection: ...


@overload
def clone_collection(type_: Literal["chapter_feed"], /) -> ChapterFeed: ...


@overload
def clone_collection(type_: Literal["covers"], /) -> CoverCollection: ...


@overload
def clone_collection(type_: Literal["custom_lists"], /) -> CustomListCollection: ...


@overload
def clone_collection(type_: Literal["legacy_mapping"], /) -> LegacyMappingCollection: ...


@overload
def clone_collection(type_: Literal["manga"], /) -> MangaCollection: ...


@overload
def clone_collection(type_: Literal["manga_relations"], /) -> MangaRelationCollection: ...


@overload
def clone_collection(type_: Literal["reports"], /) -> ReportCollection: ...


@overload
def clone_collection(type_: Literal["scanlator_groups"], /) -> ScanlatorGroupCollection: ...


@overload
def clone_collection(type_: Literal["users"], /) -> UserCollection: ...


@overload
def clone_collection(type_: Literal["user_reports"], /) -> UserReportCollection: ...


def clone_collection(type_: CollectionType, /) -> BaseCollection[Any]:
    path = PATH / f"{type_}.json"
    if type_ == "authors":
        author_payload: GetMultiAuthorResponse = json.load(path.open())
        authors = [Author(HTTP, item) for item in author_payload["data"]]
        return AuthorCollection(HTTP, author_payload, authors=authors)
    if type_ == "chapter_feed":
        chapter_payload: GetMultiChapterResponse = json.load(path.open())
        chapters = [Chapter(HTTP, item) for item in chapter_payload["data"]]
        return ChapterFeed(HTTP, chapter_payload, chapters=chapters)
    if type_ == "covers":
        cover_payload: GetMultiCoverResponse = json.load(path.open())
        covers = [Cover(HTTP, item) for item in cover_payload["data"]]
        return CoverCollection(HTTP, cover_payload, covers=covers)
    if type_ == "custom_lists":
        custom_list_payload: GetMultiCustomListResponse = json.load(path.open())
        custom_lists = [CustomList(HTTP, item) for item in custom_list_payload["data"]]
        return CustomListCollection(HTTP, custom_list_payload, lists=custom_lists)
    if type_ == "legacy_mapping":
        mapping_payload: GetLegacyMappingResponse = json.load(path.open())
        mappings = [LegacyItem(HTTP, item) for item in mapping_payload["data"]]
        return LegacyMappingCollection(HTTP, mapping_payload, mappings=mappings)
    if type_ == "manga":
        manga_payload: MangaSearchResponse = json.load(path.open())
        manga: list[Manga] = [Manga(HTTP, item) for item in manga_payload["data"]]
        return MangaCollection(HTTP, manga_payload, manga=manga)
    if type_ == "manga_relations":
        relation_payload: MangaRelationResponse = json.load(path.open())
        parent_id: str = ""
        manga_relation: list[MangaRelation] = [MangaRelation(HTTP, parent_id, item) for item in relation_payload["data"]]
        return MangaRelationCollection(HTTP, relation_payload, relations=manga_relation)
    if type_ == "scanlator_groups":
        group_payload: GetMultiScanlationGroupResponse = json.load(path.open())
        groups = [ScanlatorGroup(HTTP, item) for item in group_payload["data"]]
        return ScanlatorGroupCollection(HTTP, group_payload, groups=groups)
    if type_ == "users":
        user_payload: GetMultiUserResponse = json.load(path.open())
        users = [User(HTTP, item) for item in user_payload["data"]]
        return UserCollection(HTTP, user_payload, users=users)

    raise RuntimeError("Unreachable code in testing suite.")


class TestCollection:
    def test_author_collection(self) -> None:
        path = PATH / "authors.json"
        payload: GetMultiAuthorResponse = json.load(path.open())
        collection = clone_collection("authors")

        assert collection.limit == payload["limit"]
        assert collection.total == payload["total"]
        assert collection.offset == payload["offset"]
        assert len(collection.items) == len(payload["data"])

    def test_chapter_feed(self) -> None:
        path = PATH / "chapter_feed.json"
        payload: GetMultiChapterResponse = json.load(path.open())
        collection = clone_collection("chapter_feed")

        assert collection.limit == payload["limit"]
        assert collection.total == payload["total"]
        assert collection.offset == payload["offset"]
        assert len(collection.items) == len(payload["data"])

    def test_cover_collection(self) -> None:
        path = PATH / "covers.json"
        payload: GetMultiCoverResponse = json.load(path.open())
        collection = clone_collection("covers")

        assert collection.limit == payload["limit"]
        assert collection.total == payload["total"]
        assert collection.offset == payload["offset"]
        assert len(collection.items) == len(payload["data"])

    def test_custom_list_collection(self) -> None:
        path = PATH / "custom_lists.json"
        payload: GetMultiCustomListResponse = json.load(path.open())
        collection = clone_collection("custom_lists")

        assert collection.limit == payload["limit"]
        assert collection.total == payload["total"]
        assert collection.offset == payload["offset"]
        assert len(collection.items) == len(payload["data"])

    def test_legacy_mapping_collection(self) -> None:
        path = PATH / "legacy_mapping.json"
        payload: GetLegacyMappingResponse = json.load(path.open())
        collection = clone_collection("legacy_mapping")

        assert collection.limit == payload["limit"]
        assert collection.total == payload["total"]
        assert collection.offset == payload["offset"]
        assert len(collection.items) == len(payload["data"])

    def test_manga_collection(self) -> None:
        path = PATH / "manga.json"
        payload: MangaSearchResponse = json.load(path.open())
        collection = clone_collection("manga")

        assert collection.limit == payload["limit"]
        assert collection.total == payload["total"]
        assert collection.offset == payload["offset"]
        assert len(collection.items) == len(payload["data"])

    def test_manga_relation_collection(self) -> None:
        path = PATH / "manga_relations.json"
        payload: MangaRelationResponse = json.load(path.open())
        collection = clone_collection("manga_relations")

        assert collection.limit == payload["limit"]
        assert collection.total == payload["total"]
        assert collection.offset == payload["offset"]
        assert len(collection.items) == len(payload["data"])

    def test_scanlator_group_collection(self) -> None:
        path = PATH / "scanlator_groups.json"
        payload: GetMultiScanlationGroupResponse = json.load(path.open())
        collection = clone_collection("scanlator_groups")

        assert collection.limit == payload["limit"]
        assert collection.total == payload["total"]
        assert collection.offset == payload["offset"]
        assert len(collection.items) == len(payload["data"])

    def test_user_collection(self) -> None:
        path = PATH / "users.json"
        payload: GetMultiUserResponse = json.load(path.open())
        collection = clone_collection("users")

        assert collection.limit == payload["limit"]
        assert collection.total == payload["total"]
        assert collection.offset == payload["offset"]
        assert len(collection.items) == len(payload["data"])



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import argparse
import importlib.metadata
import platform
import sys

import aiohttp

import hondana


def show_version() -> None:
    version_info = sys.version_info
    md_version_info = hondana.version_info

    entries: list[str] = [
        f"- Python v{version_info.major}.{version_info.minor}.{version_info.micro}-{version_info.releaselevel}",
        f"- Hondana v{md_version_info.major}.{md_version_info.minor}.{md_version_info.micro}-{md_version_info.releaselevel}",
    ]

    if md_version_info.releaselevel != "final":
        try:
            version = importlib.metadata.version("Hondana")
        except importlib.metadata.PackageNotFoundError:
            pass
        else:
            entries.append(f"    - Hondana pkg_resources: v{version}")

    entries.append(f" - aiohttp {aiohttp.__version__}")
    uname = platform.uname()
    entries.append(f"- System Info: {uname.system} {uname.release} {uname.version}")

    print("\n".join(entries))  # noqa: T201 # this is intended


def parse_args() -> tuple[argparse.ArgumentParser, argparse.Namespace]:
    parser = argparse.ArgumentParser(prog="hondana", description="Tools for helping with Hondana")
    parser.add_argument("-v", "--version", action="store_true", help="shows the wrapper version")

    parser.set_defaults(func=core)

    return parser, parser.parse_args()


def core(_: argparse.ArgumentParser, args: argparse.Namespace) -> None:
    if args.version:
        show_version()


def main() -> None:
    parser, args = parse_args()
    args.func(parser, args)


main()



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import json
import operator
import pathlib
from copy import deepcopy
from typing import TYPE_CHECKING, Any

from hondana.manga import Manga
from hondana.tags import QueryTags, Tag

if TYPE_CHECKING:
    from collections.abc import Mapping

    from hondana.http import HTTPClient
    from hondana.relationship import Relationship
    from hondana.types_.manga import GetMangaResponse


PATH: pathlib.Path = pathlib.Path(__file__).parent / "payloads" / "manga.json"

PAYLOAD: GetMangaResponse = json.load(PATH.open(encoding="utf-8"))
HTTP: HTTPClient = object()  # pyright: ignore[reportAssignmentType] # this is just for test purposes.


def clone_tags() -> list[Tag]:
    t = deepcopy(PAYLOAD)
    return Manga(HTTP, t["data"]).tags


class TestTags:
    def test_ids(self) -> None:
        tags = clone_tags()

        tag_ids = [t.id for t in tags]

        assert len(tag_ids) == len(PAYLOAD["data"]["attributes"]["tags"])

    def test_tag_names(self) -> None:
        tags = clone_tags()

        raw_tags: list[str] = []
        for item in PAYLOAD["data"]["attributes"]["tags"]:
            raw_tags.extend(str(sub_key) for sub_key in item["attributes"]["name"].values())

        for tag, raw_tag in zip(
            sorted(tags, key=lambda t: t.name),
            sorted(raw_tags, key=lambda t: t),
            strict=False,
        ):
            assert tag.name == raw_tag

    def test_tag_relationships(
        self,
    ) -> None:  # currently, tags have no relationships, but even so.
        tags = clone_tags()

        tag_rels: list[Relationship] = [r for tag in tags for r in tag.relationships]

        raw_rels: list[Mapping[str, Any]] = [
            obj for rel in PAYLOAD["data"]["attributes"]["tags"] for obj in rel["relationships"]
        ]

        for a, b in zip(
            sorted(tag_rels, key=lambda r: r.id),
            sorted(raw_rels, key=operator.itemgetter("id")),
            strict=True,
        ):
            assert a == b

    def test_tag_descriptions(
        self,
    ) -> None:  # currently, tags have no descriptions, but even so.
        tags = clone_tags()
        raw_tags = PAYLOAD["data"]["attributes"]["tags"]

        for tag, raw_tag in zip(
            sorted(tags, key=lambda t: t.id),
            sorted(raw_tags, key=operator.itemgetter("id")),
            strict=True,
        ):
            description = tag._description  # pyright: ignore[reportPrivateUsage] # sorry, need this for test purposes
            raw_descriptions = raw_tag["attributes"]["description"]
            raw_descriptions = raw_descriptions or {}
            raw_description = dict(raw_descriptions)
            assert description == raw_description


class TestQueryTags:
    def test_query_tags(self) -> None:
        tags = QueryTags("Comedy", "Genderswap", "Romance", mode="OR")
        assert tags.mode == "OR"
        assert tags.tags == [
            "4d32cc48-9f00-4cca-9b5a-a839f0764984",
            "2bd2e8d0-f146-434a-9b51-fc9ff2c5fe6a",
            "423e2eae-a7a2-4a8b-ac03-a8351462d71d",
        ]

        tags = QueryTags("Comedy", "Mecha", mode="AND")
        assert tags.mode == "AND"
        assert tags.tags == [
            "4d32cc48-9f00-4cca-9b5a-a839f0764984",
            "50880a9d-5440-4732-9afb-8f457127e836",
        ]

        try:
            tags = QueryTags("SomethingElse", mode="OR")
        except ValueError:
            pass
        else:
            raise AssertionError("Tags failed to fail.")



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Literal

from .artist import Artist
from .author import Author
from .collections import ChapterFeed, MangaRelationCollection
from .cover import Cover
from .enums import ContentRating, MangaRelationType, MangaState, MangaStatus, PublicationDemographic, ReadingStatus
from .forums import MangaComments
from .query import ArtistIncludes, AuthorIncludes, ChapterIncludes, CoverIncludes, FeedOrderQuery, MangaIncludes
from .tags import Tag
from .utils import MISSING, RelationshipResolver, cached_slot_property, require_authentication

if TYPE_CHECKING:
    from .http import HTTPClient
    from .tags import QueryTags
    from .types_ import manga
    from .types_.artist import ArtistResponse
    from .types_.author import AuthorResponse
    from .types_.common import LanguageCode, LocalizedString
    from .types_.cover import CoverResponse
    from .types_.manga import MangaResponse
    from .types_.relationship import RelationshipResponse
    from .types_.statistics import (
        BatchStatisticsResponse,
        CommentMetaData,
        MangaStatisticsResponse,
        PersonalMangaRatingsResponse,
    )


__all__ = (
    "Manga",
    "MangaRating",
    "MangaRelation",
    "MangaStatistics",
)


class Manga:
    """A class representing a Manga returned from the MangaDex API.

    Attributes
    ----------
    id: :class:`str`
        The UUID associated to this manga.
    relation_type: Optional[:class:`~hondana.MangaRelationType`]
        The type of relation this is, to the parent manga requested.
        Only available when :meth:`get_related_manga` is called.
    alternate_titles: :class:`~hondana.types_.common.LocalizedString`
        The alternative title mapping for the Manga.
        i.e. ``{"en": "Some Other Title"}``
    locked: :class:`bool`
        If the Manga is considered 'locked' or not in the API.
    links: :class:`~hondana.types_.manga.MangaLinks`
        The mapping of links the API has attributed to the Manga.
        (see: https://api.mangadex.org/docs.html#section/Static-data/Manga-links-data)
    original_language: :class:`str`
        The language code for the original language of the Manga.
    last_volume: Optional[:class:`str`]
        The last volume attributed to the manga, if any.
    last_chapter: Optional[:class:`str`]
        The last chapter attributed to the manga, if any.
    publication_demographic: Optional[:class:`~hondana.PublicationDemographic`]
        The attributed publication demographic(s) for the manga, if any.
    year: Optional[:class:`int`]
        The year the manga was release, if the key exists.
    content_rating: Optional[:class:`~hondana.ContentRating`]
        The content rating attributed to the manga, if any.
    chapter_numbers_reset_on_new_volume: :class:`bool`
        Whether the chapter numbers will reset on a new volume or not.
    latest_uploaded_chapter: :class:`str`
        The ID of the latest uploaded chapter of this manga.
    state: Optional[:class:`~hondana.MangaState`]
        The publication state of the Manga.
    stats: Optional[:class:`~hondana.MangaStatistics`]
        The statistics of the manga.
    version: :class:`int`
        The version revision of this manga.


    .. note::
        The :attr:`stats` is only populated after :meth:`~hondana.Manga.get_statistics` is called.
    """

    __slots__ = (
        "__artists",
        "__authors",
        "__cover",
        "__related_manga",
        "_artist_relationships",
        "_attributes",
        "_author_relationships",
        "_cover_relationship",
        "_created_at",
        "_cs_tags",
        "_data",
        "_description",
        "_http",
        "_related_manga_relationships",
        "_tags",
        "_title",
        "_updated_at",
        "alternate_titles",
        "available_translated_languages",
        "chapter_numbers_reset_on_new_volume",
        "content_rating",
        "id",
        "last_chapter",
        "last_volume",
        "latest_uploaded_chapter",
        "links",
        "locked",
        "original_language",
        "publication_demographic",
        "relation_type",
        "state",
        "stats",
        "status",
        "version",
        "year",
    )

    def __init__(self, http: HTTPClient, payload: manga.MangaResponse) -> None:
        self._http = http
        self._data = payload
        relationships: list[RelationshipResponse] = self._data.pop("relationships", [])
        self._attributes = payload["attributes"]
        self.id: str = payload["id"]
        self._title: LocalizedString = self._attributes["title"]
        self._description: LocalizedString = self._attributes["description"]
        related = payload.get("related", None)
        self.relation_type: MangaRelationType | None = MangaRelationType(related) if related else None
        self.alternate_titles: LocalizedString = {k: v for item in self._attributes["altTitles"] for k, v in item.items()}  # pyright: ignore[reportAttributeAccessIssue]  # TypedDict.items() is weird
        self.locked: bool = self._attributes.get("isLocked", False)
        self.links: manga.MangaLinks = self._attributes["links"]
        self.original_language: str = self._attributes["originalLanguage"]
        self.last_volume: str | None = self._attributes["lastVolume"]
        self.last_chapter: str | None = self._attributes["lastChapter"]
        self.publication_demographic: PublicationDemographic | None = (
            PublicationDemographic(self._attributes["publicationDemographic"])
            if self._attributes["publicationDemographic"]
            else None
        )
        self.status: MangaStatus | None = MangaStatus(self._attributes["status"])
        self.year: int | None = self._attributes["year"]
        self.content_rating: ContentRating | None = (
            ContentRating(self._attributes["contentRating"]) if self._attributes["contentRating"] else None
        )
        self.chapter_numbers_reset_on_new_volume: bool = self._attributes["chapterNumbersResetOnNewVolume"]
        self.available_translated_languages: list[LanguageCode] = self._attributes["availableTranslatedLanguages"]
        self.latest_uploaded_chapter: str = self._attributes["latestUploadedChapter"]
        self.state: MangaState | None = MangaState(self._attributes["state"]) if self._attributes["state"] else None
        self.stats: MangaStatistics | None = None
        self.version: int = self._attributes["version"]
        self._tags = self._attributes["tags"]
        self._created_at = self._attributes["createdAt"]
        self._updated_at = self._attributes["updatedAt"]
        self._author_relationships: list[AuthorResponse] = RelationshipResolver["AuthorResponse"](
            relationships,
            "author",
        ).resolve()
        self._artist_relationships: list[ArtistResponse] = RelationshipResolver["ArtistResponse"](
            relationships,
            "artist",
        ).resolve()
        self._related_manga_relationships: list[MangaResponse] = RelationshipResolver["MangaResponse"](
            relationships,
            "manga",
        ).resolve()
        self._cover_relationship: CoverResponse | None = RelationshipResolver["CoverResponse"](
            relationships,
            "cover_art",
        ).pop(with_fallback=True)
        self.__authors: list[Author] | None = None
        self.__artists: list[Artist] | None = None
        self.__cover: Cover | None = None
        self.__related_manga: list[Manga] | None = None

    def __repr__(self) -> str:
        return f"<Manga id={self.id!r} title={self.title!r}>"

    def __str__(self) -> str:
        return self.title

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, (Manga, MangaRelation)) and self.id == other.id

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    @property
    def url(self) -> str:
        """The URL to this manga.

        Returns
        -------
        :class:`str`
            The URL of the manga.
        """
        return f"https://mangadex.org/title/{self.id}"

    @property
    def title(self) -> str:
        """The manga's title.

        Returns
        -------
        :class:`str`
            The title of the manga, defaults to the ``en`` key in the titles attribute of the response.
            Attempts to fall back to the manga's default language, and failing that it will
            fall back to the next available key if ``en`` or default language key is not present.
        """
        title = self._title.get("en")
        if title is None:
            key = next(iter(self._title))
            return self._title.get(self.original_language, self._title[key])  # pyright: ignore[reportUnknownArgumentType] # this is safe since the key is from the dict

        return title

    @property
    def description(self) -> str | None:
        """The manga's description/synopsis.

        Returns
        -------
        Optional[:class:`str`]
            The description of the manga, defaults to the ``en`` key in the titles.
            Falls back to the next available key if ``en`` is not present. If there is
            no description, returns None.
        """
        desc = self._description.get("en")
        if desc is None:
            if not self._description:
                return None
            key = next(iter(self._description))
            return self._description[key]  # pyright: ignore[reportUnknownVariableType] # this is safe since the key is from the dict
        return desc

    @property
    def raw_description(self) -> LocalizedString:
        """The raw description attribute from the manga's payload from the API.

        Returns
        -------
        :class:`~hondana.types_.common.LocalizedString`
            The raw object from the manga's api response payload.
            Provides no formatting on its own.
            Consider :meth:`~hondana.Manga.description` or :meth:`~hondana.Manga.localised_description` instead.
        """
        return self._description

    @property
    def alt_titles(self) -> LocalizedString:
        """The raw alt_titles attribute from the manga's payload from the API.

        Returns
        -------
        :class:`hondana.types_.common.LocalizedString`
            The raw object from the manga's payload.
            Provides no formatting on its own. Consider :meth:`~hondana.Manga.localised_title` instead.
        """
        return self.alternate_titles

    @property
    def created_at(self) -> datetime.datetime:
        """The date this manga was created.

        Returns
        -------
        :class:`datetime.datetime`
            The UTC datetime of when this manga was created.
        """
        return datetime.datetime.fromisoformat(self._created_at)

    @property
    def updated_at(self) -> datetime.datetime:
        """The date this manga was last updated.

        Returns
        -------
        :class:`datetime.datetime`
            The UTC datetime of when this manga was last updated.
        """
        return datetime.datetime.fromisoformat(self._updated_at)

    @cached_slot_property("_cs_tags")
    def tags(self) -> list[Tag]:
        """The tags of this Manga.

        Returns
        -------
        List[:class:`~hondana.Tag`]
            The list of tags that this manga is associated with.
        """
        return [Tag(item) for item in self._tags]

    @property
    def artists(self) -> list[Artist] | None:
        """The artists of the parent Manga.

        .. note::
            If the parent manga was **not** requested with the "artist" `includes[]` query parameter
            then this method will return ``None``.

        Returns
        -------
        Optional[List[:class:`~hondana.Artist`]]
            The artists associated with this Manga.
        """
        if self.__artists is not None:
            return self.__artists

        if not self._artist_relationships:
            return None

        formatted: list[Artist] = [
            Artist(self._http, artist) for artist in self._artist_relationships if "attributes" in artist
        ]

        if not formatted:
            return None

        self.__artists = formatted
        return self.__artists

    @artists.setter
    def artists(self, value: list[Artist]) -> None:
        self.__artists = value

    @property
    def authors(self) -> list[Author] | None:
        """The artists of the parent Manga.

        .. note::
            If the parent manga was **not** requested with the "artist" `includes[]` query parameter
            then this method will return ``None``.

        Returns
        -------
        Optional[List[:class:`~hondana.Author`]]
            The artists associated with this Manga.

        """
        if self.__authors is not None:
            return self.__authors

        if not self._author_relationships:
            return None

        formatted: list[Author] = [
            Author(self._http, author) for author in self._author_relationships if "attributes" in author
        ]

        if not formatted:
            return None

        self.__authors = formatted
        return self.__authors

    @authors.setter
    def authors(self, value: list[Author]) -> None:
        self.__authors = value

    @property
    def cover(self) -> Cover | None:
        """The cover of the manga.

        .. note::
            If the parent manga was **not** requested with the "cover" `includes[]` query parameter
            then this method will return ``None``.

        Returns
        -------
        Optional[:class:`~hondana.Cover`]
            The cover of the manga.
        """
        if self.__cover is not None:
            return self.__cover

        if not self._cover_relationship:
            return None

        if "attributes" in self._cover_relationship:
            self.__cover = Cover(self._http, self._cover_relationship)
            return self.__cover

        return None

    @cover.setter
    def cover(self, value: Cover) -> None:
        self.__cover = value

    @property
    def related_manga(self) -> list[Manga] | None:
        """The related manga of the parent Manga.

        Returns
        -------
        Optional[List[:class:`~hondana.Manga`]]
            The related manga of the parent manga.
        """
        if self.__related_manga is not None:
            return self.__related_manga

        if not self._related_manga_relationships:
            return None

        formatted: list[Manga] = [
            self.__class__(self._http, item) for item in self._related_manga_relationships if "attributes" in item
        ]

        if not formatted:
            return None

        self.__related_manga = formatted
        return self.__related_manga

    @related_manga.setter
    def related_manga(self, value: list[Manga]) -> None:
        self.__related_manga = value

    async def get_artists(self) -> list[Artist] | None:
        """|coro|

        This method will return the artists of the manga and caches the response.


        .. note::
            If the parent manga was requested with the "artist" `includes[]` query parameter,
            then this method will not make extra API calls to retrieve the artist data.

        Returns
        -------
        Optional[List[:class:`~hondana.Author`]]
            The artists of the manga.
        """
        if self.artists is not None:
            return self.artists

        if not self._artist_relationships:
            return None

        ids = [r["id"] for r in self._artist_relationships]

        formatted: list[Artist] = []
        for item in ids:
            data = await self._http.get_artist(item, includes=ArtistIncludes())
            formatted.append(Artist(self._http, data["data"]))

        if not formatted:
            return None

        self.artists = formatted
        return formatted

    async def get_authors(self) -> list[Author] | None:
        """|coro|

        This method will return the authors of the manga and caches the response.


        .. note::
            If the parent manga was requested with the "author" `includes[]` query parameter,
            then this method will not make extra API calls to retrieve the author data.

        Returns
        -------
        Optional[List[:class:`~hondana.Author`]]
            The authors of the manga.
        """
        if self.authors is not None:
            return self.authors

        if not self._author_relationships:
            return None

        ids = [r["id"] for r in self._related_manga_relationships]

        formatted: list[Author] = []
        for item in ids:
            data = await self._http.get_author(item, includes=AuthorIncludes())
            formatted.append(Author(self._http, data["data"]))

        if not formatted:
            return None

        self.authors = formatted
        return formatted

    async def get_cover(self) -> Cover | None:
        """|coro|

        This method will return the cover URL of the parent Manga if it exists and caches the response.

        Returns
        -------
        Optional[:class:`~hondana.Cover`]
            The Cover associated with this Manga.
        """
        if self.cover is not None:
            return self.cover

        if not self._cover_relationship:
            return None

        data = await self._http.get_cover(self._cover_relationship["id"], includes=CoverIncludes())
        self.cover = Cover(self._http, data["data"])
        return self.cover

    def cover_url(self, *, size: Literal[256, 512] | None = None) -> str | None:
        """Method to return a direct url to the cover art of the parent Manga.

        If the manga was requested without the ``"cover_art"`` includes[] parameters, then this method will return ``None``.

        Returns
        -------
        Optional[:class:`str`]
            The cover url, if present in the underlying manga details.


        .. note::
            For a more stable cover return, try :meth:`~hondana.Manga.get_cover`
        """
        if not self.cover:
            return None

        return self.cover.url(size, parent_id=self.id)

    async def get_related_manga(self, *, limit: int = 100, offset: int = 0) -> list[Manga] | None:
        """|coro|

        This method will return all the related manga and cache their response.

        Parameters
        ----------
        limit: :class:`int`
            The amount of manga to fetch. Defaults to ``100``.
        offset: :class:`int`
            The pagination offset. Defaults to ``0``.


        .. note::
            If the parent manga was requested with the "manga" `includes[]` query parameter,
            then this method will not make extra API calls to retrieve manga data.

        Returns
        -------
        Optional[List[:class:`~hondana.Manga`]]
            The related manga of the parent.
        """
        if self.related_manga is not None:
            return self.related_manga

        if not self._related_manga_relationships:
            return None

        ids = [r["id"] for r in self._related_manga_relationships]

        data = await self._http.manga_list(
            limit=limit,
            offset=offset,
            title=None,
            author_or_artist=None,
            authors=None,
            artists=None,
            year=None,
            included_tags=None,
            excluded_tags=None,
            status=None,
            original_language=None,
            excluded_original_language=None,
            available_translated_language=None,
            publication_demographic=None,
            ids=ids,
            content_rating=None,
            created_at_since=None,
            updated_at_since=None,
            order=None,
            includes=MangaIncludes(),
            has_available_chapters=None,
            has_unavailable_chapters=None,
            group=None,
        )

        ret: list[Manga] = [Manga(self._http, item) for item in data["data"]]
        if not ret:
            return None

        self.related_manga = ret
        return self.related_manga

    @require_authentication
    async def update(
        self,
        *,
        title: LocalizedString | None = None,
        alt_titles: list[LocalizedString] | None = None,
        description: LocalizedString | None = None,
        authors: list[str] | None = None,
        artists: list[str] | None = None,
        links: manga.MangaLinks | None = None,
        original_language: str | None = None,
        last_volume: str | None = MISSING,
        last_chapter: str | None = MISSING,
        publication_demographic: PublicationDemographic = MISSING,
        status: MangaStatus | None = None,
        year: int | None = MISSING,
        content_rating: ContentRating | None = None,
        tags: QueryTags | None = None,
        primary_cover: str | None = MISSING,
        version: int,
    ) -> Manga:
        """|coro|

        This method will update the current Manga within the MangaDex API.

        Parameters
        ----------
        title: Optional[:class:`~hondana.types_.common.LocalizedString`]
            The manga titles in the format of ``language_key: title``
            i.e. ``{"en": "Some Manga Title"}``
        alt_titles: Optional[List[:class:`~hondana.types_.common.LocalizedString`]]
            The alternative titles in the format of ``language_key: title``
            i.e. ``[{"en": "Some Other Title"}, {"fr": "Un Autre Titre"}]``
        description: Optional[:class:`~hondana.types_.common.LocalizedString`]
            The manga description in the format of ``language_key: description``
            i.e. ``{"en": "My amazing manga where x y z happens"}``
        authors: Optional[List[:class:`str`]]
            The list of author UUIDs to credit to this manga.
        artists: Optional[List[:class:`str`]]
            The list of artist UUIDs to credit to this manga.
        links: Optional[:class:`~hondana.types_.manga.MangaLinks`]
            The links relevant to the manga.
            See here for more details: https://api.mangadex.org/docs.html#section/Static-data/Manga-links-data
        original_language: Optional[:class:`str`]
            The language key for the original language of the manga.
        last_volume: Optional[:class:`str`]
            The last volume to attribute to this manga.
        last_chapter: Optional[:class:`str`]
            The last chapter to attribute to this manga.
        publication_demographic: Optional[:class:`~hondana.PublicationDemographic`]
            The target publication demographic of this manga.
        status: Optional[:class:`~hondana.MangaStatus`]
            The status of the manga.
        year: Optional[:class:`int`]
            The release year of the manga.
        content_rating: Optional[:class:`~hondana.ContentRating`]
            The content rating of the manga.
        tags: Optional[:class:`QueryTags`]
            The QueryTags instance for the list of tags to attribute to this manga.
        primmary_cover: Optional[:class:`str`]
            The UUID representing the cover that should show for this manga as it's primary.
        version: :class:`int`
            The revision version of this manga.


        .. note::
            The ``last_volume``, ``last_chapter``, ``publication_demographic``, ``year`` and ``primary_cover`` parameters
            are nullable in the API. Pass ``None`` explicitly to do so.

        Raises
        ------
        BadRequest
            The query parameters were not valid.
        Forbidden
            The update returned an error due to authentication failure.
        NotFound
            The specified manga does not exist.

        Returns
        -------
        :class:`~hondana.Manga`
            The manga that was returned after creation.
        """  # noqa: DOC502 # raised in method call
        data = await self._http.update_manga(
            self.id,
            title=title,
            alt_titles=alt_titles,
            description=description,
            authors=authors,
            artists=artists,
            links=links,
            original_language=original_language,
            last_volume=last_volume,
            last_chapter=last_chapter,
            publication_demographic=publication_demographic,
            status=status,
            year=year,
            content_rating=content_rating,
            tags=tags,
            primary_cover=primary_cover,
            version=version,
        )

        return self.__class__(self._http, data["data"])

    @require_authentication
    async def delete(self) -> None:
        """|coro|

        This method will delete a Manga from the MangaDex API.

        Raises
        ------
        Forbidden
            The update returned an error due to authentication failure.
        NotFound
            The specified manga does not exist.
        """  # noqa: DOC502 # raised in method call
        await self._http.delete_manga(self.id)

    @require_authentication
    async def unfollow(self) -> None:
        """|coro|

        This method will unfollow the current Manga for the logged-in user in the MangaDex API.

        Raises
        ------
        Forbidden
            The request returned an error due to authentication failure.
        NotFound
            The specified manga does not exist.
        """  # noqa: DOC502 # raised in method call
        await self._http.unfollow_manga(self.id)

    @require_authentication
    async def follow(self, *, set_status: bool = True, status: ReadingStatus = ReadingStatus.reading) -> None:
        """|coro|

        This method will follow the current Manga for the logged-in user in the MangaDex API.

        Parameters
        ----------
        set_status: :class:`bool`
            Whether to set the reading status of the manga you follow.
            Due to the current MangaDex infrastructure, not setting a
            status will cause the manga to not show up in your lists.
            Defaults to ``True``
        status: :class:`~hondana.ReadingStatus`
            The status to apply to the newly followed manga.
            Irrelevant if ``set_status`` is ``False``.

        Raises
        ------
        Forbidden
            The request returned an error due to authentication failure.
        NotFound
            The specified manga does not exist.
        """  # noqa: DOC502 # raised in method call
        await self._http.follow_manga(self.id)
        if set_status:
            await self.update_reading_status(status=status)

    def localised_title(self, language_code: LanguageCode, /) -> str | None:
        """Method to return the current manga's title in the provided language code.

        Falling back to the :attr:`title`.

        Aliased to :meth:`~Manga.localized_title`

        Parameters
        ----------
        language_code: :class:`~hondana.types_.common.LanguageCode`
            The language code to attempt to return the manga name in.

        Returns
        -------
        Optional[:class:`str`]
            The manga name in the provided language, if found.
        """
        return self.alternate_titles.get(language_code, self.title)

    localized_title = localised_title

    def localised_description(self, language_code: LanguageCode, /) -> str | None:
        """Method to return the current manga's description in the provided language code.

        Falling back to the :attr:`description`.

        Aliased to :meth:`~Manga.localized_description`

        Parameters
        ----------
        language_code: :class:`~hondana.types_.common.LanguageCode`
            The language code to attempt to return the manga description in.

        Returns
        -------
        Optional[:class:`str`]
            The manga description in the provided language, if found.
        """
        return self._description.get(language_code, self.description)

    localized_description = localised_description

    async def feed(
        self,
        *,
        limit: int | None = 100,
        offset: int = 0,
        translated_language: list[LanguageCode] | None = None,
        original_language: list[LanguageCode] | None = None,
        excluded_original_language: list[LanguageCode] | None = None,
        content_rating: list[ContentRating] | None = None,
        excluded_groups: list[str] | None = None,
        excluded_uploaders: list[str] | None = None,
        include_future_updates: bool | None = None,
        created_at_since: datetime.datetime | None = None,
        updated_at_since: datetime.datetime | None = None,
        published_at_since: datetime.datetime | None = None,
        order: FeedOrderQuery | None = None,
        includes: ChapterIncludes | None = None,
        include_empty_pages: bool | None = None,
        include_future_publish_at: bool | None = None,
        include_external_url: bool | None = None,
        include_unavailable: bool | None = None,
    ) -> ChapterFeed:
        """|coro|

        This method returns the current manga's chapter feed.

        Parameters
        ----------
        limit: Optional[:class:`int`]
            Defaults to 100. The maximum amount of chapters to return-in the response.
        offset: :class:`int`
            Defaults to 0. The pagination offset for the request.
        translated_language: List[:class:`str`]
            A list of language codes to filter the returned chapters with.
        original_language: List[:class:`~hondana.types_.common.LanguageCode`]
            A list of language codes to filter the original language of the returned chapters with.
        excluded_original_language: List[:class:`~hondana.types_.common.LanguageCode`]
            A list of language codes to negate filter the original language of the returned chapters with.
        content_rating: Optional[List[:class:`~hondana.ContentRating`]]
            The content rating to filter the feed by.
        excluded_groups: Optional[List[:class:`str`]]
            The list of scanlator groups to exclude from the response.
        excluded_uploaders: Optional[List[:class:`str`]]
            The list of uploaders to exclude from the response.
        include_future_updates: Optional[:class:`bool`]
            Whether to include future chapters in this feed. Defaults to ``True`` API side.
        created_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their creation date.
        updated_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their updated at date.
        published_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their published at date.
        order: Optional[:class:`~hondana.query.FeedOrderQuery`]
            A query parameter to choose how the responses are ordered.
            i.e. ``{"chapters": "desc"}``
        includes: Optional[:class:`~hondana.query.ChapterIncludes`]
            The list of options to include increased payloads for per chapter.
            Defaults to these values.
        include_empty_pages: Optional[:class:`bool`]
            Whether to show chapters with no pages available.
        include_future_publish_at: Optional[:class:`bool`]
            Whether to show chapters with a publishAt value set in the future.
        include_external_url: Optional[:class:`bool`]
            Whether to show chapters that have an external URL attached to them.
        include_unavailable: Optional[:class:`bool`]
            Whether to show chapters that are marked as unavailable.


        .. note::
            Passing ``None`` to ``limit`` will attempt to retrieve all items in the chapter feed.

        Raises
        ------
        BadRequest
            The query parameters were malformed.

        Returns
        -------
        :class:`~hondana.ChapterFeed`
            Returns a collection of chapters.
        """  # noqa: DOC502 # raised in method call
        inner_limit = limit or 100

        chapters: list[Chapter] = []
        while True:
            data = await self._http.manga_feed(
                self.id,
                limit=inner_limit,
                offset=offset,
                translated_language=translated_language,
                original_language=original_language,
                excluded_original_language=excluded_original_language,
                content_rating=content_rating,
                excluded_groups=excluded_groups,
                excluded_uploaders=excluded_uploaders,
                include_future_updates=include_future_updates,
                created_at_since=created_at_since,
                updated_at_since=updated_at_since,
                published_at_since=published_at_since,
                order=order,
                includes=includes or ChapterIncludes(),
                include_empty_pages=include_empty_pages,
                include_future_publish_at=include_future_publish_at,
                include_external_url=include_external_url,
                include_unavailable=include_unavailable,
            )

            from .chapter import Chapter  # noqa: PLC0415 # cyclic import cheat

            chapters.extend([Chapter(self._http, item) for item in data["data"]])

            offset += inner_limit
            if not data["data"] or offset >= 10_000 or limit is not None:
                break

        return ChapterFeed(self._http, data, chapters)

    @require_authentication
    async def update_read_markers(self) -> manga.MangaReadMarkersResponse:
        """|coro|

        This method will return the read chapters of the current manga.

        Returns
        -------
        :class:`hondana.types_.manga.MangaReadMarkersResponse`
            The raw payload of the API.
            Contains a list of read chapter UUIDs.
        """
        return await self._http.manga_read_markers([self.id], grouped=False)

    @require_authentication
    async def bulk_update_read_markers(
        self,
        *,
        update_history: bool = True,
        read_chapters: list[str] | None,
        unread_chapters: list[str] | None,
    ) -> None:
        """|coro|

        This method will batch update your read chapters for a given Manga.

        Parameters
        ----------
        update_history: :class:`bool`
            Whether to show this chapter in the authenticated user's read history.
            Defaults to ``True``.
        read_chapters: Optional[List[:class:`str`]]
            The read chapters for this Manga.
        unread_chapters: Optional[List[:class:`str`]]
            The unread chapters for this Manga.

        Raises
        ------
        TypeError
            You must provide one or both of the parameters `read_chapters` and/or `unread_chapters`.
        """
        if read_chapters or unread_chapters:
            await self._http.manga_read_markers_batch(
                self.id,
                update_history=update_history,
                read_chapters=read_chapters,
                unread_chapters=unread_chapters,
            )
            return

        msg = "You must provide either `read_chapters` and/or `unread_chapters` to this method."
        raise TypeError(msg)

    @require_authentication
    async def get_reading_status(self) -> manga.MangaSingleReadingStatusResponse:
        """|coro|

        This method will return the current reading status for the current manga.

        Raises
        ------
        Forbidden
            You are not authenticated to perform this action.
        NotFound
            The specified manga does not exist, likely due to an incorrect ID.

        Returns
        -------
        :class:`~hondana.types_.manga.MangaSingleReadingStatusResponse`
            The raw payload from the API response.
        """  # noqa: DOC502 # raised in method call
        return await self._http.get_manga_reading_status(self.id)

    @require_authentication
    async def update_reading_status(self, *, status: ReadingStatus) -> None:
        """|coro|

        This method will update your current reading status for the current manga.

        Parameters
        ----------
        status: :class:`~hondana.ReadingStatus`
            The reading status you wish to update this manga with.


        .. note::
            Leaving ``status`` as the default will remove the manga reading status from the API.
            Please provide a value if you do not wish for this to happen.

        Raises
        ------
        BadRequest
            The query parameters were invalid.
        NotFound
            The specified manga cannot be found, likely due to incorrect ID.
        """  # noqa: DOC502 # raised in method call
        await self._http.update_manga_reading_status(self.id, status=status)

    async def get_volumes_and_chapters(
        self,
        *,
        translated_language: list[LanguageCode] | None = None,
        groups: list[str] | None = None,
    ) -> manga.GetMangaVolumesAndChaptersResponse:
        """|coro|

        This endpoint returns the raw relational mapping of a manga's volumes and chapters.

        Parameters
        ----------
        translated_language: Optional[List[:class:`~hondana.types_.common.LanguageCode`]]
            The list of language codes you want to limit the search to.
        groups: Optional[List[:class:`str`]]
            A list of scanlator groups to filter the results by.

        Returns
        -------
        :class:`~hondana.types_.manga.GetMangaVolumesAndChaptersResponse`
            The raw payload from mangadex. There is no guarantee of the keys here.
        """
        return await self._http.get_manga_volumes_and_chapters(
            manga_id=self.id,
            translated_language=translated_language,
            groups=groups,
        )

    @require_authentication
    async def add_to_custom_list(self, *, custom_list_id: str) -> None:
        """|coro|

        This method will add the current manga to the specified custom list.

        Parameters
        ----------
        custom_list_id: :class:`str`
            The UUID associated with the custom list you wish to add the manga to.

        Raises
        ------
        Forbidden
            You are not authorised to add manga to this custom list.
        NotFound
            The specified manga or specified custom list are not found, likely due to an incorrect UUID.
        """  # noqa: DOC502 # raised in method call
        await self._http.add_manga_to_custom_list(custom_list_id, manga_id=self.id)

    @require_authentication
    async def remove_from_custom_list(self, *, custom_list_id: str) -> None:
        """|coro|

        This method will remove the current manga from the specified custom list.

        Parameters
        ----------
        custom_list_id: :class:`str`
            THe UUID associated with the custom list you wish to add the manga to.

        Raises
        ------
        Forbidden
            You are not authorised to remove a manga from the specified custom list.
        NotFound
            The specified manga or specified custom list are not found, likely due to an incorrect UUID.
        """  # noqa: DOC502 # raised in method call
        await self._http.remove_manga_from_custom_list(custom_list_id, manga_id=self.id)

    async def get_chapters(
        self,
        *,
        limit: int | None = 10,
        offset: int = 0,
        ids: list[str] | None = None,
        title: str | None = None,
        groups: list[str] | None = None,
        uploader: str | None = None,
        volumes: list[str] | None = None,
        chapter: list[str] | None = None,
        translated_language: list[LanguageCode] | None = None,
        original_language: list[LanguageCode] | None = None,
        excluded_original_language: list[LanguageCode] | None = None,
        content_rating: list[ContentRating] | None = None,
        excluded_groups: list[str] | None = None,
        excluded_uploaders: list[str] | None = None,
        include_future_updates: bool | None = None,
        include_empty_pages: bool | None = None,
        include_future_publish_at: bool | None = None,
        include_external_url: bool | None = None,
        include_unavailable: bool | None = None,
        created_at_since: datetime.datetime | None = None,
        updated_at_since: datetime.datetime | None = None,
        published_at_since: datetime.datetime | None = None,
        order: FeedOrderQuery | None = None,
        includes: ChapterIncludes | None = None,
    ) -> ChapterFeed:
        """|coro|

        This method will return a list of published chapters.

        Parameters
        ----------
        limit: Optional[:class:`int`]
            Defaults to 100. This specifies the amount of chapters to return in one request.
        offset: :class:`int`
            Defaults to 0. This specifies the pagination offset.
        ids: Optional[List[:class:`str`]]
            The list of chapter UUIDs to filter the request with.
        title: Optional[:class:`str`]
            The chapter title query to limit the request with.
        groups: Optional[List[:class:`str`]]
            The scanlation group UUID(s) to limit the request with.
        uploader: Optional[:class:`str`]
            The uploader UUID to limit the request with.
        volumes: Optional[Union[:class:`str`, List[:class:`str`]]]
            The volume UUID or UUIDs to limit the request with.
        chapter: Optional[Union[:class:`str`, List[:class:`str`]]]
            The chapter UUID or UUIDs to limit the request with.
        translated_language: Optional[List[:class:`~hondana.types_.common.LanguageCode`]]
            The list of languages codes to filter the request with.
        original_language: Optional[List[:class:`~hondana.types_.common.LanguageCode`]]
            The list of languages to specifically target in the request.
        excluded_original_language: Optional[List[:class:`~hondana.types_.common.LanguageCode`]]
            The list of original languages to exclude from the request.
        content_rating: Optional[List[:class:`~hondana.ContentRating`]]
            The content rating to filter the feed by.
        excluded_groups: Optional[List[:class:`str`]]
            The list of scanlator groups to exclude from the response.
        excluded_uploaders: Optional[List[:class:`str`]]
            The list of uploaders to exclude from the response.
        include_future_updates: Optional[:class:`bool`]
            Whether to include future chapters in this feed. Defaults to ``True`` API side.
        include_empty_pages: Optional[:class:`bool`]
            Whether to include chapters that have no recorded pages.
        include_future_publish_at: Optional[:class:`bool`]
            Whether to include chapters that have their publish time set to a time in the future.
        include_external_url: Optional[:class:`bool`]
            Whether to include chapters that have an external url set.
        include_unavailable: Optional[:class:`bool`]
            Whether to include chapters that are marked as unavailable.
        created_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their creation date.
        updated_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their updated at date.
        published_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their published at date.
        order: Optional[:class:`~hondana.query.FeedOrderQuery`]
            A query parameter to choose how the responses are ordered.
            i.e. ``{"chapters": "desc"}``
        includes: Optional[:class:`~hondana.query.ChapterIncludes`]
            The list of options to include increased payloads for per chapter.
            Defaults to all values.


        .. note::
            Passing ``None`` to ``limit`` will attempt to retrieve all items in the chapter feed.

        .. note::
            If `order` is not specified then the API will return results first based on their creation date,
            which could lead to unexpected results.

        Raises
        ------
        BadRequest
            The query parameters were malformed

        Returns
        -------
        :class:`~hondana.ChapterFeed`
            Returns a collection of chapters.
        """  # noqa: DOC502 # raised in method call
        inner_limit = limit or 10

        chapters: list[Chapter] = []
        while True:
            data = await self._http.chapter_list(
                limit=inner_limit,
                offset=offset,
                manga=self.id,
                ids=ids,
                title=title,
                groups=groups,
                uploader=uploader,
                volume=volumes,
                chapter=chapter,
                translated_language=translated_language,
                original_language=original_language,
                excluded_original_language=excluded_original_language,
                content_rating=content_rating,
                excluded_groups=excluded_groups,
                excluded_uploaders=excluded_uploaders,
                include_future_updates=include_future_updates,
                include_empty_pages=include_empty_pages,
                include_future_publish_at=include_future_publish_at,
                include_external_url=include_external_url,
                include_unavailable=include_unavailable,
                created_at_since=created_at_since,
                updated_at_since=updated_at_since,
                published_at_since=published_at_since,
                order=order,
                includes=includes or ChapterIncludes(),
            )
            from .chapter import Chapter  # noqa: PLC0415 # cyclic import cheat

            chapters.extend([Chapter(self._http, item) for item in data["data"]])

            offset += inner_limit
            if not data["data"] or offset >= 10_000 or limit is not None:
                break

        return ChapterFeed(self._http, data, chapters)

    async def get_draft(self) -> Manga:
        """|coro|

        This method will return a manga draft from MangaDex.

        Returns
        -------
        :class:`~hondana.Manga`
            The Manga returned from the API.
        """
        data = await self._http.get_manga_draft(self.id)
        return self.__class__(self._http, data["data"])

    async def submit_draft(self, *, version: int) -> Manga:
        """|coro|

        This method will submit a draft for a manga.

        Parameters
        ----------
        version: :class:`int`
            The version of the manga we're attributing this submission to.

        Raises
        ------
        BadRequest
            The request parameters were incorrect or malformed.
        Forbidden
            You are not authorised to perform this action.
        NotFound
            The manga was not found.

        Returns
        -------
        :class:`~hondana.Manga`
        """  # noqa: DOC502 # raised in method call
        data = await self._http.submit_manga_draft(self.id, version=version)
        return self.__class__(self._http, data["data"])

    async def get_relations(self, *, includes: MangaIncludes | None = None) -> MangaRelationCollection:
        """|coro|

        This method will return a list of all relations to a given manga.

        Parameters
        ----------
        includes: Optional[:class:`~hondana.query.MangaIncludes`]
            The optional parameters for expanded requests to the API.
            Defaults to all possible expansions.

        Raises
        ------
        BadRequest
            The manga ID passed is malformed

        Returns
        -------
        :class:`~hondana.MangaRelationCollection`
        """  # noqa: DOC502 # raised in method call
        data = await self._http.get_manga_relation_list(self.id, includes=includes or MangaIncludes())
        fmt = [MangaRelation(self._http, self.id, item) for item in data["data"]]
        return MangaRelationCollection(self._http, data, fmt)

    @require_authentication
    async def upload_cover(
        self,
        *,
        cover: bytes,
        volume: str | None = None,
        description: str,
        locale: LanguageCode | None = None,
    ) -> Cover:
        """|coro|

        This method will upload a cover to the MangaDex API.

        Parameters
        ----------
        cover: :class:`bytes`
            THe raw bytes of the image.
        volume: Optional[:class:`str`]
            The volume this cover relates to.
        description: :class:`str`
            The description of this cover.
        locale: Optional[:class:`~hondana.types_.common.LanguageCode`]
            The locale of this cover.

        Raises
        ------
        BadRequest
            The volume parameter was malformed or the file was a bad format.
        Forbidden
            You are not permitted for this action.

        Returns
        -------
        :class:`~hondana.Cover`
        """  # noqa: DOC502 # raised in method call
        data = await self._http.upload_cover(self.id, cover=cover, volume=volume, description=description, locale=locale)
        return Cover(self._http, data["data"])

    @require_authentication
    async def create_relation(self, *, target_manga: str, relation_type: MangaRelationType) -> MangaRelation:
        """|coro|

        This method will create a manga relation.

        Parameters
        ----------
        target_manga: :class:`str`
            The manga ID of the related manga.
        relation_type: :class:`~hondana.MangaRelationType`

        Raises
        ------
        BadRequest
            The parameters were malformed
        Forbidden
            You are not authorised for this action.

        Returns
        -------
        :class:`~hondana.MangaRelation`
        """  # noqa: DOC502 # raised in method call
        data = await self._http.create_manga_relation(self.id, target_manga=target_manga, relation_type=relation_type)
        return MangaRelation(self._http, self.id, data["data"])

    @require_authentication
    async def delete_relation(self, relation_id: str, /) -> None:
        """|coro|

        This method will delete a manga relation.

        Parameters
        ----------
        relation_id: :class:`str`
            The ID of the related manga.
        """
        await self._http.delete_manga_relation(self.id, relation_id)

    @require_authentication
    async def set_rating(self, *, rating: int) -> None:
        """|coro|

        This method will set your rating on the manga.
        This method **overwrites** your previous set rating, if any.

        Parameters
        ----------
        rating: :class:`int`
            The rating value, between 0 and 10.

        Raises
        ------
        Forbidden
            The request returned a response due to authentication failure.
        NotFound
            The specified manga UUID was not found or does not exist.
        """  # noqa: DOC502 # raised in method call
        await self._http.set_manga_rating(self.id, rating=rating)

    @require_authentication
    async def delete_rating(self) -> None:
        """|coro|

        This method will delete your set rating on the manga.

        Raises
        ------
        Forbidden
            The request returned a response due to authentication failure.
        NotFound
            The specified manga UUID was not found or does not exist.
        """  # noqa: DOC502 # raised in method call
        await self._http.delete_manga_rating(self.id)

    async def get_statistics(self) -> MangaStatistics:
        """|coro|

        This method will fetch statistics on the current manga, and cache them as the :attr:`stats`

        Returns
        -------
        :class:`~hondana.MangaStatistics`
        """
        data = await self._http.get_manga_statistics(self.id, None)

        key = next(iter(data["statistics"]))
        stats = MangaStatistics(self._http, self.id, data["statistics"][key])

        self.stats = stats
        return self.stats


class MangaRelation:
    """A class representing a MangaRelation returned from the MangaDex API.

    Attributes
    ----------
    source_manga_id: :class:`str`
        The UUID associated to the parent manga of this relation.
    id: :class:`str`
        The UUID associated to this manga relation.
    version: :class:`int`
        The version revision of this manga relation.
    relation_type: :class:`~hondana.MangaRelationType`
        The type of relationship to the source manga.
    """

    __slots__ = (
        "_attributes",
        "_data",
        "_http",
        "_relationships",
        "id",
        "relation_type",
        "source_manga_id",
        "version",
    )

    def __init__(self, http: HTTPClient, parent_id: str, payload: manga.MangaRelation, /) -> None:
        self._http = http
        self._data = payload
        self._attributes = self._data["attributes"]
        self._relationships = self._data.pop("relationships", [])
        self.source_manga_id: str = parent_id
        self.id: str = self._data["id"]
        self.version: int = self._attributes["version"]
        self.relation_type: MangaRelationType = MangaRelationType(self._attributes["relation"])

    def __repr__(self) -> str:
        return f"<MangaRelation id={self.id!r} source_manga_id={self.source_manga_id!r}>"

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, (Manga, MangaRelation)) and (self.id == other.id or self.source_manga_id == other.id)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)


class MangaStatistics:
    """
    A small object to house manga statistics.

    Attributes
    ----------
    follows: :class:`int`
        The number of follows this manga has.
    parent_id: :class:`str`
        The manga these statistics belong to.
    average: Optional[:class:`float`]
        The average mean score of the manga ratings.
    bayesian: Optional[:class:`float`]
        The bayesian average score of the manga ratings.
    distribution: Optional[Dict[:class:`str`, :class:`int`]]
        The scoring distribution of the manga.
        Keys are 1-10 and values are total amount of ratings per key.


    .. note::
        The :attr:`distribution` attribute will be None unless this object was
        created with :meth:`hondana.Client.get_manga_statistics` or :meth:`hondana.Manga.get_statistics`
    """

    __slots__ = (
        "_comments",
        "_cs_comments",
        "_data",
        "_http",
        "_rating",
        "average",
        "bayesian",
        "distribution",
        "follows",
        "parent_id",
        "unavilable_chapter_count",
    )

    def __init__(self, http: HTTPClient, parent_id: str, payload: MangaStatisticsResponse | BatchStatisticsResponse) -> None:
        self._http: HTTPClient = http
        self._data = payload
        self._rating = payload["rating"]
        self._comments: CommentMetaData | None = payload.get("comments")
        self.follows: int = payload["follows"]
        self.parent_id: str = parent_id
        self.average: float | None = self._rating["average"]
        self.bayesian: float | None = self._rating["bayesian"]
        self.distribution: dict[str, int] | None = self._rating.get("distribution")
        self.unavilable_chapter_count: int = self._data.get("unavailableChapterCount", 0)

    def __repr__(self) -> str:
        return f"<MangaStatistics for={self.parent_id!r}>"

    @cached_slot_property("_cs_comments")
    def comments(self) -> MangaComments | None:
        """
        Returns the comments helper object if the target object has the relevant data (has comments, basically).

        Returns
        -------
        Optional[:class:`hondana.MangaComments`]
        """
        if self._comments:
            return MangaComments(self._http, self._comments, self.parent_id)

        return None


class MangaRating:
    """
    A small object to encompass your personal manga ratings.

    Attributes
    ----------
    parent_id: :class:`str`
        The parent manga this rating belongs to.
    rating: :class:`int`
        Your personal rating for this manga, between 1 and 10.
    created_at: :class:`datetime.datetime`
        When you created the rating, as a UTC aware datetime.
    """

    __slots__ = (
        "_data",
        "_http",
        "created_at",
        "parent_id",
        "rating",
    )

    def __init__(self, http: HTTPClient, parent_id: str, payload: PersonalMangaRatingsResponse) -> None:
        self._http = http
        self._data = payload
        self.parent_id: str = parent_id
        self.rating: int = self._data["rating"]
        self.created_at: datetime.datetime = datetime.datetime.fromisoformat(self._data["createdAt"])

    def __repr__(self) -> str:
        return f"<MangaRating parent_id={self.parent_id!r}>"



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
import json
import pathlib
from copy import deepcopy
from typing import TYPE_CHECKING, Literal, overload

from hondana.enums import MangaRelationType
from hondana.manga import Manga, MangaRating, MangaRelation, MangaStatistics
from hondana.utils import RelationshipResolver, to_snake_case

if TYPE_CHECKING:
    from hondana.http import HTTPClient
    from hondana.types_.artist import ArtistResponse
    from hondana.types_.author import AuthorResponse
    from hondana.types_.common import LocalizedString
    from hondana.types_.cover import CoverResponse
    from hondana.types_.manga import GetMangaResponse, MangaRelationResponse, MangaResponse
    from hondana.types_.statistics import GetMangaStatisticsResponse, GetPersonalMangaRatingsResponse


PATH: pathlib.Path = pathlib.Path(__file__).parent / "payloads" / "manga.json"
RELATIONS_PATH: pathlib.Path = pathlib.Path(__file__).parent / "payloads" / "manga_relations.json"
STATISTICS_PATH: pathlib.Path = pathlib.Path(__file__).parent / "payloads" / "manga_statistics.json"
RATING_PATH: pathlib.Path = pathlib.Path(__file__).parent / "payloads" / "manga_ratings.json"

PAYLOAD: GetMangaResponse = json.load(PATH.open(encoding="utf-8"))
RELATIONS_PAYLOAD: MangaRelationResponse = json.load(RELATIONS_PATH.open(encoding="utf-8"))
STATISTICS_PAYLOAD: GetMangaStatisticsResponse = json.load(STATISTICS_PATH.open(encoding="utf-8"))
RATING_PAYLOAD: GetPersonalMangaRatingsResponse = json.load(RATING_PATH.open(encoding="utf-8"))
HTTP: HTTPClient = object()  # pyright: ignore[reportAssignmentType] # this is just for test purposes.


@overload
def clone_manga(type_: Literal["manga"]) -> Manga: ...


@overload
def clone_manga(type_: Literal["relation"]) -> MangaRelation: ...


@overload
def clone_manga(type_: Literal["stats"]) -> MangaStatistics: ...


@overload
def clone_manga(type_: Literal["rating"]) -> MangaRating: ...


def clone_manga(
    type_: Literal["manga", "relation", "stats", "rating"] = "manga",
) -> Manga | MangaRelation | MangaStatistics | MangaRating:
    if type_ == "manga":
        t = deepcopy(PAYLOAD)
        return Manga(HTTP, t["data"])
    if type_ == "relation":
        t = deepcopy(RELATIONS_PAYLOAD)
        return MangaRelation(HTTP, PAYLOAD["data"]["id"], t["data"][0])
    if type_ == "stats":
        t = deepcopy(STATISTICS_PAYLOAD)
        key = next(iter(t["statistics"]))
        return MangaStatistics(HTTP, PAYLOAD["data"]["id"], t["statistics"][key])
    t = deepcopy(RATING_PAYLOAD)
    key = next(iter(t["ratings"]))
    return MangaRating(HTTP, PAYLOAD["data"]["id"], t["ratings"][key])


class TestManga:
    def test_id(self) -> None:
        manga = clone_manga("manga")
        assert manga.id == PAYLOAD["data"]["id"]

    def test_attributes(self) -> None:
        manga = clone_manga("manga")

        for item in PAYLOAD["data"]["attributes"]:
            if item == "altTitles":  # special case sane attribute renaming
                item = "alternateTitles"  # noqa: PLW2901  # renaming raw payload items
            elif item == "isLocked":  # special case sane attribute renaming
                item = "locked"  # noqa: PLW2901  # renaming raw payload items
            assert hasattr(manga, to_snake_case(item))

    def test_relationship_length(self) -> None:
        manga = clone_manga("manga")
        assert manga.artists is not None
        assert manga.authors is not None
        assert manga.related_manga is not None
        assert manga.cover is not None

        obj_len = len(manga.artists) + len(manga.authors) + len(manga.related_manga) + 1  # cover

        assert "relationships" in PAYLOAD["data"]

        assert obj_len == len(PAYLOAD["data"]["relationships"])

    def test_cache_slot_property(self) -> None:
        manga = clone_manga("manga")

        assert not hasattr(manga, "_cs_tags")

        _ = manga.tags

        assert hasattr(manga, "_cs_tags")

    def test_artists_property(self) -> None:
        manga = clone_manga("manga")

        assert manga.artists is not None
        assert "relationships" in PAYLOAD["data"]
        artist_rels = RelationshipResolver["ArtistResponse"](PAYLOAD["data"]["relationships"], "artist").resolve()

        assert len(manga.artists) == len(artist_rels)

    def test_authors_property(self) -> None:
        manga = clone_manga("manga")

        assert manga.authors is not None
        assert "relationships" in PAYLOAD["data"]
        author_rels = RelationshipResolver["AuthorResponse"](PAYLOAD["data"]["relationships"], "author").resolve()

        assert len(manga.authors) == len(author_rels)

    def test_cover_property(self) -> None:
        manga = clone_manga("manga")

        assert manga.cover is not None

        assert "relationships" in PAYLOAD["data"]
        cover_rel = RelationshipResolver["CoverResponse"](PAYLOAD["data"]["relationships"], "cover_art").pop()
        assert cover_rel is not None

        assert manga.cover.id == cover_rel["id"]
        assert manga.cover_url() == f"https://uploads.mangadex.org/covers/{manga.id}/{cover_rel['attributes']['fileName']}"
        assert (
            manga.cover_url(size=256)
            == f"https://uploads.mangadex.org/covers/{manga.id}/{cover_rel['attributes']['fileName']}.256.jpg"
        )
        assert (
            manga.cover_url(size=512)
            == f"https://uploads.mangadex.org/covers/{manga.id}/{cover_rel['attributes']['fileName']}.512.jpg"
        )

    def test_related_manga_property(self) -> None:
        manga = clone_manga("manga")

        assert manga.related_manga is not None

        assert "relationships" in PAYLOAD["data"]
        related_keys = RelationshipResolver["MangaResponse"](PAYLOAD["data"]["relationships"], "manga").resolve()
        assert bool(related_keys)

        assert len(manga.related_manga) == len(related_keys)

    def test_alt_title(self) -> None:
        manga = clone_manga("manga")

        alt_titles = PAYLOAD["data"]["attributes"]["altTitles"]

        fmt: LocalizedString = {k: v for obj in alt_titles for k, v in obj.items()}  # pyright: ignore[reportAssignmentType] # TypedDict.items() is weird

        for code, title in fmt.items():
            assert manga.alternate_titles.get(code) == title

    def test_localized_title(self) -> None:
        manga = clone_manga("manga")

        assert manga.title == next(iter(PAYLOAD["data"]["attributes"]["title"].values()))  # why did I need to do this

    def test_localized_description(self) -> None:
        manga = clone_manga("manga")

        for key, value in manga._description.items():  # pyright: ignore[reportPrivateUsage] # sorry, need this for test purposes
            assert manga.localized_description(key) == value  # pyright: ignore[reportArgumentType] # can't narrow strings

    def test_date_attributes(self) -> None:
        manga = clone_manga("manga")

        assert manga.created_at == datetime.datetime.fromisoformat(PAYLOAD["data"]["attributes"]["createdAt"])
        assert manga.updated_at == datetime.datetime.fromisoformat(PAYLOAD["data"]["attributes"]["updatedAt"])


class TestMangaRelation:
    def test_id(self) -> None:
        manga = clone_manga("relation")

        assert manga.source_manga_id == PAYLOAD["data"]["id"]
        assert manga.id == RELATIONS_PAYLOAD["data"][0]["id"]

    def test_type(self) -> None:
        manga = clone_manga("relation")

        assert manga.relation_type.value == RELATIONS_PAYLOAD["data"][0]["attributes"]["relation"]
        assert manga.relation_type is MangaRelationType(RELATIONS_PAYLOAD["data"][0]["attributes"]["relation"])


class TestMangaStatistics:
    def test_id(self) -> None:
        manga = clone_manga("stats")

        assert manga.parent_id == PAYLOAD["data"]["id"]

    def test_bookmarks(self) -> None:
        manga = clone_manga("stats")

        key = next(iter(STATISTICS_PAYLOAD["statistics"]))
        assert manga.follows == STATISTICS_PAYLOAD["statistics"][key]["follows"]

    def test_average(self) -> None:
        manga = clone_manga("stats")

        key = next(iter(STATISTICS_PAYLOAD["statistics"]))
        assert manga.average == STATISTICS_PAYLOAD["statistics"][key]["rating"]["average"]

    def test_bayesian(self) -> None:
        manga = clone_manga("stats")

        key = next(iter(STATISTICS_PAYLOAD["statistics"]))
        assert manga.bayesian == STATISTICS_PAYLOAD["statistics"][key]["rating"]["bayesian"]


class TestMangaRating:
    def test_id(self) -> None:
        manga = clone_manga("rating")

        assert manga.parent_id == PAYLOAD["data"]["id"]

    def test_rating(self) -> None:
        manga = clone_manga("rating")

        key = next(iter(RATING_PAYLOAD["ratings"]))
        assert manga.rating == RATING_PAYLOAD["ratings"][key]["rating"]



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from .forums import ScanlatorGroupComments
from .utils import MISSING, RelationshipResolver, cached_slot_property, deprecated, iso_to_delta, require_authentication

if TYPE_CHECKING:
    from .http import HTTPClient
    from .types_.common import LanguageCode
    from .types_.relationship import RelationshipResponse
    from .types_.scanlator_group import ScanlationGroupResponse
    from .types_.statistics import CommentMetaData, StatisticsCommentsResponse
    from .types_.user import UserResponse
    from .user import User  # noqa: TC004

__all__ = (
    "ScanlatorGroup",
    "ScanlatorGroupStatistics",
)


class ScanlatorGroup:
    """
    A class representing a Scanlator Group from the MangaDex API.

    Attributes
    ----------
    id: :class:`str`
        The UUID relating to this scanlator group.
    name: :class:`str`
        The name of the scanlator group.
    alt_names: List[:class:`str`]
        The alternate names for this group, if any.
    website: :class:`str`
        The website of this scanlator group.
    irc_server: :class:`str`
        The IRC server for this scanlator group.
    irc_channel: :class:`str`
        The IRC channel for this scanlator group.
    discord: :class:`str`
        The Discord server for this scanlator group.
    focused_languages: Optional[List[:class:`str`]]
        The scanlator group's focused languages, if any.
    contact_email: :class:`str`
        The contact email for this scanlator group.
    description: :class:`str`
        The description of this scanlator group.
    twitter: Optional[:class:`str`]
        The twitter URL of the group, if any.
    manga_updates: Optional[:class:`str`]
        The URL where the group posts their updates, if any.
    locked: :class:`bool`
        If this scanlator group is considered 'locked' or not.
    official: :class:`bool`
        Whether the group is official or not.
    ex_licensed: :class:`bool`
        TBC.
    verified: :class:`bool`
        Whether this group is verified or not.
    inactive: :class:`bool`
        Whether this group is inactive or not.
    version: :class:`int`
        The version revision of this scanlator group.
    """

    __slots__ = (
        "__leader",
        "__members",
        "_attributes",
        "_created_at",
        "_data",
        "_http",
        "_leader_relationship",
        "_member_relationships",
        "_publish_delay",
        "_stats",
        "_updated_at",
        "alt_names",
        "contact_email",
        "description",
        "discord",
        "ex_licensed",
        "focused_languages",
        "id",
        "inactive",
        "irc_channel",
        "irc_server",
        "locked",
        "manga_updates",
        "name",
        "official",
        "twitter",
        "verified",
        "version",
        "website",
    )

    def __init__(self, http: HTTPClient, payload: ScanlationGroupResponse) -> None:
        self._http = http
        self._data = payload
        self._attributes = self._data["attributes"]
        self.id: str = self._data["id"]
        relationships: list[RelationshipResponse] = self._data.pop("relationships", [])
        self.name: str = self._attributes["name"]
        self.alt_names: list[str] = self._attributes["altNames"]
        self.website: str | None = self._attributes["website"]
        self.irc_server: str | None = self._attributes["ircServer"]
        self.irc_channel: str | None = self._attributes["ircServer"]
        self.discord: str | None = self._attributes["discord"]
        self.focused_languages: list[str] | None = self._attributes.get("focusedLanguages")
        self.contact_email: str | None = self._attributes["contactEmail"]
        self.description: str | None = self._attributes["description"]
        self.twitter: str | None = self._attributes["twitter"]
        self.manga_updates: str | None = self._attributes["mangaUpdates"]
        self.locked: bool = self._attributes.get("locked", False)
        self.official: bool = self._attributes["official"]
        self.verified: bool = self._attributes["verified"]
        self.inactive: bool = self._attributes["inactive"]
        self.ex_licensed: bool = self._attributes.get("exLicensed", False)
        self.version: int = self._attributes["version"]
        self._created_at = self._attributes["createdAt"]
        self._updated_at = self._attributes["updatedAt"]
        self._publish_delay: str = self._attributes["publishDelay"]
        self._stats: ScanlatorGroupStatistics | None = None
        self._leader_relationship: UserResponse | None = RelationshipResolver["UserResponse"](
            relationships,
            "leader",
        ).pop(with_fallback=True)
        self._member_relationships: list[UserResponse] = RelationshipResolver["UserResponse"](
            relationships,
            "member",
        ).resolve()
        self.__leader: User | None = None
        self.__members: list[User] | None = None

    def __repr__(self) -> str:
        return f"<ScanlatorGroup id={self.id!r} name={self.name!r}>"

    def __str__(self) -> str:
        return self.name

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ScanlatorGroup) and self.id == other.id

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    @property
    def stats(self) -> ScanlatorGroupStatistics | None:
        """Returns the statistics object of the scanlator group if it was fetched and cached.

        Returns
        -------
        Optional[:class:`hondana.ScanlatorGroupStatistics`]
        """
        return self._stats

    @property
    def created_at(self) -> datetime.datetime:
        """
        Returns the time when the ScanlatorGroup was created.

        Returns
        -------
        :class:`datetime.datetime`
        """
        return datetime.datetime.fromisoformat(self._created_at)

    @property
    def updated_at(self) -> datetime.datetime:
        """
        Returns the time when the ScanlatorGroup was last updated.

        Returns
        -------
        :class:`datetime.datetime`
        """
        return datetime.datetime.fromisoformat(self._updated_at)

    @property
    def url(self) -> str:
        """The URL to this scanlator group.

        Returns
        -------
        :class:`str`
            The URL of the scanlator group.
        """
        return f"https://mangadex.org/group/{self.id}"

    @property
    def publish_delay(self) -> datetime.timedelta | None:
        """The publishing delay of this scanlation group.

        Returns
        -------
        Optional[:class:`datetime.timedelta`]
            The default timedelta offset at which this group releases their chapters, if any.
        """
        if self._publish_delay:
            return iso_to_delta(self._publish_delay)

        return None

    @property
    def leader(self) -> User | None:
        """The leader of this scanlation group, if any.

        Returns
        -------
        Optional[:class:`~hondana.User`]
            The leader of the scanlation group.
        """
        if self.__leader is not None:
            return self.__leader

        if not self._leader_relationship:
            return None

        if "attributes" in self._leader_relationship:
            from .user import User  # noqa: PLC0415 # cyclic import cheat

            user = User(self._http, self._leader_relationship)
            self.__leader = user
            return self.__leader

        return None

    @leader.setter
    def leader(self, other: User) -> None:
        self.__leader = other

    @property
    def members(self) -> list[User] | None:
        """The members of this scanlation group, if any.

        Returns
        -------
        Optional[List[:class:`~hondana.User`]]
            The members of the scanlation group.
        """
        if self.__members is not None:
            return self.__members

        if not self._member_relationships:
            return None

        fmt: list[User] = []

        for relationship in self._member_relationships:
            from .user import User  # noqa: PLC0415 # cyclic import cheat

            if "attributes" in relationship:
                fmt.append(User(self._http, relationship))

        if not fmt:
            return None

        self.__members = fmt
        return self.__members

    async def get_leader(self) -> User | None:
        """|coro|

        This method will return the User representing the leader of the scanlation group.

        .. note::
            If the ScanlatorGroup was requested with the `leader` includes, then this method will not make an API call.


        .. note::
            If this object was created as part of another object's ``includes`` then this will return None.
            This is due to having no relationship data.


        Returns
        -------
        Optional[User]
            The leader of the ScanlatorGroup, if present.
        """
        if self.leader is not None:
            return self.leader

        if not self._leader_relationship:
            return None

        leader_id = self._leader_relationship["id"]
        data = await self._http.get_user(leader_id)

        from .user import User  # noqa: PLC0415 # cyclic import cheat

        leader = User(self._http, data["data"])
        self.__leader = leader
        return self.__leader

    async def get_members(self) -> list[User] | None:
        """|coro|

        This method will return a list of members of the scanlation group, if present.

        .. note::
            If the ScanlatorGroup was requested with the ``member`` includes, this method will not make API calls.


        .. note::
            If this object was created as part of another object's ``includes`` then this will return None.
            This is due to having no relationship data.

        Returns
        -------
        Optional[List[User]]
            The list of members of the scanlation group.
        """
        if self.members is not None:
            return self.members

        if not self._member_relationships:
            return None

        ids = [r["id"] for r in self._member_relationships]

        data = await self._http.user_list(limit=100, offset=0, ids=ids, username=None, order=None)

        self.__members = [User(self._http, payload) for payload in data["data"]]
        return self.__members

    @require_authentication
    async def delete(self) -> None:
        """|coro|

        This method will delete the current scanlation group.

        Raises
        ------
        Forbidden
            You are not authorized to delete this scanlation group.
        NotFound
            The scanlation group cannot be found, likely due to an incorrect ID.
        """  # noqa: DOC502 # raised in method call
        await self._http.delete_scanlation_group(self.id)

    @require_authentication
    @deprecated("bookmark")
    async def follow(self) -> None:
        """|coro|

        This method will follow the current scanlation group.

        Raises
        ------
        NotFound
            The scanlation group cannot be found, likely due to an incorrect ID.
        """  # noqa: DOC502 # raised in method call
        await self._http.follow_scanlation_group(self.id)

    bookmark = follow

    @require_authentication
    @deprecated("unbookmark")
    async def unfollow(self) -> None:
        """|coro|

        This method will unfollow the current scanlation group.

        Raises
        ------
        NotFound
            The scanlation group cannot be found, likely due to an incorrect ID.
        """  # noqa: DOC502 # raised in method call
        await self._http.unfollow_scanlation_group(self.id)

    unbookmark = unfollow

    @require_authentication
    async def update(
        self,
        *,
        name: str | None = None,
        leader: str | None = None,
        members: list[str] | None = None,
        website: str | None = MISSING,
        irc_server: str | None = MISSING,
        irc_channel: str | None = MISSING,
        discord: str | None = MISSING,
        contact_email: str | None = MISSING,
        description: str | None = MISSING,
        twitter: str | None = MISSING,
        manga_updates: str | None = MISSING,
        focused_languages: list[LanguageCode] = MISSING,
        inactive: bool | None = None,
        locked: bool | None = None,
        publish_delay: str | datetime.timedelta | None = MISSING,
        version: int,
    ) -> ScanlatorGroup:
        """|coro|

        This method will update a scanlation group within the MangaDex API.

        Parameters
        ----------
        name: Optional[:class:`str`]
            The name to update the group with.
        leader: Optional[:class:`str`]
            The UUID of the leader to update the group with.
        members: Optional[:class:`str`]
            A list of UUIDs relating to the members to update the group with.
        website: Optional[:class:`str`]
            The website to update the group with.
        irc_server: Optional[:class:`str`]
            The IRC Server to update the group with.
        irc_channel: Optional[:class:`str`]
            The IRC Channel to update the group with.
        discord: Optional[:class:`str`]
            The discord server to update the group with.
        contact_email: Optional[:class:`str`]
            The contact email to update the group with.
        description: Optional[:class:`str`]
            The new description to update the group with.
        twitter: Optional[:class:`str`]
            The new twitter url to update the group with.
        manga_updates: Optional[:class:`str`]
            The URL for this group's manga updates page, if any.
        focused_languages: Optional[List[:class:`~hondana.types_.common.LanguageCode`]]
            The new list of language codes to update the group with.
        inactive: Optional[:class:`bool`]
            If the group is inactive or not.
        locked: Optional[:class:`bool`]
            Update the lock status of this scanlator group.
        publish_delay: Optional[Union[:class:`str`, :class:`datetime.timedelta`]]
            The publishing delay to add to all group releases.
        version: :class:`int`
            The version revision of this scanlator group.


        .. note::
            The ``website``, ``irc_server``, ``irc_channel``, ``discord``, ``contact_email``, ``description``,
            ``twitter``, ``manga_updates``, ``focused_language`` and ``publish_delay``
            keys are all nullable in the API. To do so pass ``None`` explicitly to these keys.

        .. note::
            The ``publish_delay`` parameter must match the :class:`hondana.utils.MANGADEX_TIME_REGEX` pattern
            or be a valid ``datetime.timedelta``.

        Raises
        ------
        BadRequest
            The request body was malformed
        Forbidden
            You are not authorized to update this scanlation group.
        NotFound
            The passed scanlation group ID cannot be found.

        Returns
        -------
        :class:`ScanlatorGroup`
            The group returned from the API after its update.
        """  # noqa: DOC502 # raised in method call
        data = await self._http.update_scanlation_group(
            self.id,
            name=name,
            leader=leader,
            members=members,
            website=website,
            irc_server=irc_server,
            irc_channel=irc_channel,
            discord=discord,
            contact_email=contact_email,
            description=description,
            twitter=twitter,
            manga_updates=manga_updates,
            focused_languages=focused_languages,
            inactive=inactive,
            locked=locked,
            publish_delay=publish_delay,
            version=version,
        )

        return self.__class__(self._http, data["data"])

    @require_authentication
    async def get_statistics(self) -> ScanlatorGroupStatistics | None:
        """|coro|

        This method will fetch statistics on the current chapter, and cache them as the :attr:`stats`

        Returns
        -------
        :class:`~hondana.MangaStatistics`
        """
        data = await self._http.get_chapter_statistics(self.id, None)

        key = next(iter(data["statistics"]))
        stats = ScanlatorGroupStatistics(self._http, self.id, data["statistics"][key])

        self._stats = stats
        return self.stats


class ScanlatorGroupStatistics:
    """
    A small object to house scanlator group statistics.

    parent_id: :class:`str`
        The manga these statistics belong to.
    """

    __slots__ = (
        "_comments",
        "_cs_comments",
        "_data",
        "_http",
        "parent_id",
    )

    def __init__(self, http: HTTPClient, parent_id: str, payload: StatisticsCommentsResponse) -> None:
        self._http: HTTPClient = http
        self._data: StatisticsCommentsResponse = payload
        self._comments: CommentMetaData | None = payload.get("comments")
        self.parent_id: str = parent_id

    def __repr__(self) -> str:
        return f"<ChapterStatistics for={self.parent_id!r}>"

    @cached_slot_property("_cs_comments")
    def comments(self) -> ScanlatorGroupComments | None:
        """
        Returns the comments helper object if the target object has the relevant data (has comments, basically).

        Returns
        -------
        Optional[:class:`hondana.ChapterComments`]
        """
        if self._comments:
            return ScanlatorGroupComments(self._http, self._comments, self.parent_id)

        return None
